"""
`qz_curves` module
==================

"""

# Import libraries
import math as m
import numpy as np
from numba import njit, prange
from random import random

import openpile.utils.misc as misc

# SPRING FUNCTIONS --------------------------------------------

# API sand function
@njit(cache=True)
def api_sand(
    sig: float,
    delta: float,
    D: float,
    output_length: int = 7,
):
    """
    Creates the API sand Q.z curve from relevant input.

    Parameters
    ----------
    sig : float
        Vertical effective stress [unit: kPa]
    delta : float
        interface friction angle [unit: degrees]
    D : float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 8, by default 8

    Returns
    -------
    numpy 1darray
        z vector [unit: m]
    numpy 1darray
        Q vector [unit: kPa]
    """
    # cannot have less than 8
    if output_length < 8:
        output_length = 8

    # unit toe reistance [kPa]
    f = misc._Qmax_api_sand(sig, delta)

    # piecewise function
    zlist = [-0.002, 0.0, 0.002, 0.013, 0.042, 0.073, 0.100, 0.200]
    Qlist = [0.0, 0.0, 0.25, 0.50, 0.75, 0.90, 1.00, 1.00]

    # determine z vector
    z = np.array(zlist, dtype=np.float32) * D

    # define t vector
    Q = np.array(Qlist, dtype=np.float32) * f

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_Q_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = (0.1 + random() * 0.1) * D
        add_Q_values[i] = f

    z = np.append(z, add_z_values)
    Q = np.append(Q, add_Q_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    Q = Q[z_id_sorted]

    return z, Q


# API clay Q-z function
@njit(cache=True)
def api_clay(
    Su: float,
    D: float,
    output_length: int = 7,
):
    """
    Creates the API clay Q.z curve from relevant input.

    Parameters
    ----------
    Su : float
        Undrained shear strength [unit: kPa]
    D: float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 8, by default 8

    Returns
    -------
    numpy 1darray
        z vector [unit: m]
    numpy 1darray
        Q vector [unit: kPa]
    """
    # cannot have less than 8
    if output_length < 8:
        output_length = 8

    # unit toe reistance [kPa]
    f = misc._Qmax_api_clay

    # piecewise function
    zlist = [-0.002, 0.0, 0.002, 0.013, 0.042, 0.073, 0.100, 0.200]
    Qlist = [0.0, 0.0, 0.25, 0.50, 0.75, 0.90, 1.00, 1.00]

    # determine z vector
    z = np.array(zlist, dtype=np.float32) * D

    # define t vector
    Q = np.array(Qlist, dtype=np.float32) * f

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_Q_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = (0.1 + random() * 0.1) * D
        add_Q_values[i] = f

    z = np.append(z, add_z_values)
    Q = np.append(Q, add_Q_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    Q = Q[z_id_sorted]

    return z, Q



import matplotlib.pyplot as plt
import openpile.utils.py_curves as py
import pytest 
import math as m
import numpy as np

import openpile.utils.py_curves as py

@pytest.fixture
def make_pmax_api_sand():
    def make(
        sig: float,
        X: float,
        phi: float,
        D: float,
    ):
        # Calculate Pmax (regular API)
        ## Factors according to Mosher and Dawkins 2008.(regular API)
        b = 0.4
        Beta = 45 + phi / 2
        rad = m.pi / 180
        C1 = (
            (b * m.tan(phi * rad) * m.sin(Beta * rad))
            / (m.tan((Beta - phi) * rad) * m.cos((phi / 2) * rad))
            + ((m.tan(Beta * rad)) ** 2 * m.tan((phi / 2) * rad)) / (m.tan((Beta - phi) * rad))
            + b * m.tan(Beta * rad) * (m.tan(phi * rad) * m.sin(Beta * rad) - m.tan((phi / 2) * rad))
        )
        C2 = m.tan(Beta * rad) / m.tan((Beta - phi) * rad) - (m.tan((45 - phi / 2) * rad)) ** 2
        C3 = b * m.tan(phi * rad) * (m.tan(Beta * rad)) ** 4 + (m.tan((45 - phi / 2) * rad)) ** 2 * (
            (m.tan(Beta * rad)) ** 8 - 1
        )

        ## Pmax for shallow and deep zones (regular API)
        return min(C3 * sig * D, C1 * sig * X + C2 * sig * D)

    return make

@pytest.fixture
def make_pmax_api_clay():
    def make(
        sig: float,
        X: float,
        Su: float,
        D: float,
        J: float,
    ):

        return min(9*Su*X, (3 * Su + sig) * D + J * Su * X)
        
    return make


@pytest.mark.parametrize("xsigma",[0,75,200])
@pytest.mark.parametrize("xX",[0.1,5,20])
@pytest.mark.parametrize("xphi",[27,35,40])
@pytest.mark.parametrize("xD",[0.2,1.5,5.0,10.0])
@pytest.mark.parametrize("xkind",["static","cyclic"])
@pytest.mark.parametrize("xwater",[True,False])
def test_api_sand(make_pmax_api_sand,xsigma,xX, xphi, xD, xkind, xwater):
    #create spring
    y, p = py.api_sand(sig=xsigma,X=xX,phi=xphi, D=xD, kind=xkind, below_water_table=xwater)
    #helper fct
    is_sorted = lambda a: np.all(a[:-1] <= a[1:])
    #check if sorted
    assert is_sorted(y)
    assert is_sorted(p)
    #check if origin is (0,0)
    assert p[0] == 0.0
    assert y[0] == 0.0
    #check pmax
    if xkind == "cyclic":
        A = 0.9
    elif xkind=="static":
        A = max(0.9, 3 - 0.8 * xX / xD)
    assert m.isclose(p[-1], A*make_pmax_api_sand(xsigma,xX,xphi,xD), rel_tol=0.01, abs_tol=0.1)


@pytest.mark.parametrize("xsigma",[0,75,200])
@pytest.mark.parametrize("xX",[0.1,5,20])
@pytest.mark.parametrize("xSu",[0.01,50,125,300])
@pytest.mark.parametrize("xe50",[0.005,0.0125,0.03])
@pytest.mark.parametrize("xD",[0.2,1.5,5.0,10.0])
@pytest.mark.parametrize("xJ",[0.25,0.5])
@pytest.mark.parametrize("xkind",["static","cyclic"])
def test_api_clay(make_pmax_api_clay,xsigma, xX, xSu, xe50, xD, xJ, xkind):
    #create spring
    y, p = py.api_clay(sig=xsigma, X=xX, Su=xSu, eps50=xe50, D=xD, J=xJ, kind=xkind)
    #helper fct
    is_sorted = lambda a: np.all(a[:-1] <= a[1:])
    #check if sorted
    assert is_sorted(y)
    #check if origin is (0,0)
    assert p[0] == 0.0
    assert y[0] == 0.0
    #check initial stiffness
    pmax = make_pmax_api_clay(xsigma,xX,xSu,xD,xJ)
    kini = 0.23*pmax/(0.25*xe50*xD)
    assert m.isclose( (p[1]-p[0])/(y[1]-y[0]), kini, rel_tol=0.02 )
    #calculate pmax
    #check residual and max p
    Xr = max((6 * xD) / (xsigma / xX * xD / xSu + xJ), 2.5 * xD)

    factor = min(1.0,xX/Xr)
        
    if xkind=="cyclic" and xSu > 96:
        pres = 0.5*pmax*factor
        pu = 0.5*pmax
    elif xkind=="cyclic" and xSu <= 96:
        pres = 0.7185*pmax*factor
        pu = 0.7185*pmax
    elif xkind=="static":
        pres = pmax
        pu = pmax
    assert m.isclose(p[-1],pres, rel_tol=0.01, abs_tol=0.1)
    assert m.isclose(np.max(p), pu, rel_tol=0.01, abs_tol=0.1)

@pytest.mark.parametrize("xEi",[1,75,200])
@pytest.mark.parametrize("xqu",[1,75,200])
@pytest.mark.parametrize("xRQD",[0,25,50,75,100])
@pytest.mark.parametrize("xxr",[0,5,20])
@pytest.mark.parametrize("xD",[1,3,6])
def test_reese_weakrock(xEi, xqu, xRQD, xxr, xD):
    #create curve
    y, p = py.reese_weakrock(Ei=xEi, qu=xqu, RQD=xRQD, xr=xxr, D=xD)
    #helper fct
    is_sorted = lambda a: np.all(a[:-1] <= a[1:])
    #check if sorted
    assert is_sorted(y)
    #check if origin is (0,0)
    assert p[0] == 0.0
    assert y[0] == 0.0

    alpha = 1 - 2/3 * xRQD/100
    pu = min(alpha * xqu * xD * (1+1.4*xxr/xD), 5.2*alpha*xqu*xD)
    assert m.isclose(np.max(p), pu, rel_tol=0.01, abs_tol=0.1)







"""
The validation module gathers all utilised functions/class/objects serving
the purpose of validation.
"""

import numpy as np


class UserInputError(Exception):
    """Custom error for wrong inputs to openpile"""


class InvalidBoundaryConditionsError(Exception):
    """Custom error for wrong inputs to boundary conditions"""


# ----------------------------------------------------------------
#                              PILE
# ----------------------------------------------------------------


def pile_sections_must_be(obj):
    if not isinstance(obj.pile_sections, dict):
        raise UserInputError("openpile.construct.Pile.pile_sections must be a dictionary")

    if obj.kind == "Circular":
        reference_list = ["diameter", "length", "wall thickness"]
        sorted_list = list(obj.pile_sections.keys())
        sorted_list.sort()
        if sorted_list != reference_list:
            raise UserInputError(
                "openpile.construct.Pile.pile_sections must have all and only the following keys: \n - "
                + "\n - ".join(reference_list)
            )
        for idx, (_, sublist) in enumerate(obj.pile_sections.items()):
            if not isinstance(sublist, list):
                raise UserInputError(
                    "openpile.construct.Pile.pile_sections must be a dictionary of lists"
                )
            for value in sublist:
                if not isinstance(value, (int, float)):
                    raise UserInputError(
                        "values in openpile.construct.Pile.pile_sections can only be numbers"
                    )

            if idx == 0:
                reference_length = len(sublist)
            else:
                if len(sublist) != reference_length:
                    raise ValueError(
                        "length of lists in openpile.construct.Pile.pile_sections must be the same"
                    )

        for i in range(reference_length):
            if obj.pile_sections["diameter"][i] / 2 < obj.pile_sections["wall thickness"][i]:
                raise ValueError(
                    "The wall thickness cannot be larger than half the diameter of the pile"
                )


def param_must_be_type(parameter, parameter_name, parameter_type, type_name):
    if not isinstance(parameter, parameter_type):
        raise UserInputError(f"Wrong type for {parameter_name}. Please provide a {type_name}")


def str_must_be_one_of_those(param: str, param_name: str, accepted_values: list):
    param_must_be_type(param, param_name, str, "string")
    if param not in accepted_values:
        raise UserInputError(
            f"Value in {param_name} type must be one of the following: \n - "
            + "\n - ".join(accepted_values)
        )


def must_be_numbers_in_list(values, values_name):
    for value in values:
        if not isinstance(value, (float, int)):
            raise UserInputError(f"values in {values_name} can only be numbers")


def check_boundary_conditions(model):
    """
    Check if boundary conditions are satisfactory to solve the system of equations.

    Parameters
    ----------
    model: openppile.construct.Model object
        Mehs object crated from openpile
    """
    # rename vars
    restrained_dof = model.global_restrained
    loaded_dof = model.global_forces

    # count BC in [Translation over z-axis,Translation over y-axis,Rotation around x-axis]
    restrained_count_Rz = np.count_nonzero(restrained_dof["Rz"])
    restrained_count_Ty = np.count_nonzero(restrained_dof["Ty"])
    restrained_count_Tx = np.count_nonzero(restrained_dof["Tx"])
    restrained_count_total = restrained_count_Rz + restrained_count_Ty + restrained_count_Tx

    loaded_count_Rz = np.count_nonzero(loaded_dof["Mz [kNm]"])
    loaded_count_Ty = np.count_nonzero(loaded_dof["Py [kN]"])
    loaded_count_Tx = np.count_nonzero(loaded_dof["Px [kN]"])
    loaded_count_total = loaded_count_Rz + loaded_count_Ty + loaded_count_Tx

    if restrained_count_total == 0:
        raise InvalidBoundaryConditionsError("No support conditions are provided.")

    if loaded_count_total == 0:
        raise InvalidBoundaryConditionsError("No load conditions are provided.")

    if model.soil is None:
        # normally loaded beam
        if loaded_count_Tx > 0 and restrained_count_Tx > 0:
            # support in x axis is given and load over x is given --> correct BC
            pass
        elif loaded_count_Tx > 0 and restrained_count_Tx == 0:
            # support in x axis is given and load over x is given --> correct BC
            raise InvalidBoundaryConditionsError(
                "Support conditions in normal direction not provided."
            )

        # laterally-loaded beam
        if (restrained_count_Ty + restrained_count_Rz) >= 2 and loaded_count_Ty > 0:
            pass
            # support in y and z axes are given and load over y is given --> correct BC
        elif (restrained_count_Ty + restrained_count_Rz) >= 2 and loaded_count_Rz > 0:
            pass
            # support in y and z axes are given and load over z is given --> correct BC
        elif (loaded_count_Rz + loaded_count_Ty) == 0:
            # no trasnverse load --> no need to give any error
            pass
        else:
            raise InvalidBoundaryConditionsError("Support conditions against bending not provided.")
    else:
        raise InvalidBoundaryConditionsError("Soil in mesh not yet possible.")



"""
`SoilModels` module
===================

"""

# Guide to create soil models
# ---------------------------
#
#


# Import libraries
import math as m
import numpy as np
import pandas as pd
from numba import njit, prange

from typing import List, Dict, Optional, Union, Callable
from typing_extensions import Literal
from pydantic import (
    BaseModel,
    Field,
    root_validator,
    PositiveFloat,
    confloat,
    conlist,
    Extra,
)
from pydantic.dataclasses import dataclass

from openpile.core.misc import from_list2x_parse_top_bottom, var_to_str
from openpile.utils import py_curves, Hb_curves, mt_curves, Mb_curves, tz_curves


# CONSTITUTIVE MODELS CLASSES ---------------------------------


class PydanticConfigFrozen:
    arbitrary_types_allowed = True
    allow_mutation = False


class ConstitutiveModel:
    pass


class LateralModel(ConstitutiveModel):
    pass


class AxialModel(ConstitutiveModel):
    pass


@dataclass(config=PydanticConfigFrozen)
class API_clay_axial(AxialModel):

    #: undrained shear strength [kPa], if a variation in values, two values can be given.
    Su: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: t-multiplier
    t_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: z-multiplier
    z_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0
    #: Q-multiplier
    Q_multiplier: confloat(ge=0.0) = 1.0
    #: w-multiplier
    w_multiplier: confloat(gt=0.0) = 1.0

    def __str__(self):
        return f"\tAPI clay\n\tSu = {var_to_str(self.Su)} kPa"


@dataclass(config=PydanticConfigFrozen)
class Cowden_clay(LateralModel):
    """A class to establish the PISA Cowden clay model as per Byrne et al 2020 (see [BHBG20]_).

    Parameters
    ----------
    Su: float or list[top_value, bottom_value]
        Undrained shear strength. Value to range from 0 to 100 [unit: kPa]
    G0: float or list[top_value, bottom_value]
        Small-strain shear modulus [unit: kPa]
    p_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for p-values
    y_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for y-values
    m_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for m-values
    t_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for t-values

    See also
    --------
    :py:func:`openpile.utils.py_curves.cowden_clay`, :py:func:`openpile.utils.mt_curves.cowden_clay`,
    :py:func:`openpile.utils.Hb_curves.cowden_clay`, :py:func:`openpile.utils.Mb_curves.cowden_clay`


    """

    #: Undrained shear strength [kPa], if a variation in values, two values can be given.
    Su: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: small-strain shear stiffness modulus [kPa]
    G0: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: p-multiplier
    p_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: y-multiplier
    y_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0
    #: m-multiplier
    m_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: t-multiplier
    t_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0

    # spring signature which tells that API sand only has p-y curves
    # signature if of the form [p-y:True, Hb:False, m-t:False, Mb:False]
    spring_signature = np.array([True, True, True, True], dtype=bool)

    def __str__(self):
        return f"\Cowden clay (PISA)\n\tSu = {var_to_str(self.Su)} kPa.\n\tG0 = {round(self.G0/1000,1)} MPa"

    def py_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Su_t, Su_b = from_list2x_parse_top_bottom(self.Su)
        Su = Su_t + (Su_b - Su_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        y, p = py_curves.cowden_clay(
            X=X,
            Su=Su,
            G0=Gmax,
            D=D,
            output_length=output_length,
        )

        # parse multipliers and apply results
        y_mult = self.y_multiplier if isinstance(self.y_multiplier, float) else self.y_multiplier(X)
        p_mult = self.p_multiplier if isinstance(self.p_multiplier, float) else self.p_multiplier(X)

        return y * y_mult, p * p_mult

    def Hb_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Su_t, Su_b = from_list2x_parse_top_bottom(self.Su)
        Su = Su_t + (Su_b - Su_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        y, Hb = Hb_curves.cowden_clay(
            X=X,
            Su=Su,
            G0=Gmax,
            D=D,
            L=L,
            output_length=output_length,
        )

        return y, Hb

    def mt_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Su_t, Su_b = from_list2x_parse_top_bottom(self.Su)
        Su = Su_t + (Su_b - Su_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        _, p_array = py_curves.cowden_clay(
            X=X,
            Su=Su,
            G0=Gmax,
            D=D,
            output_length=output_length,
        )

        m = np.zeros((output_length, output_length), dtype=np.float32)
        t = np.zeros((output_length, output_length), dtype=np.float32)

        for count, _ in enumerate(p_array):
            t[count, :], m[count, :] = mt_curves.cowden_clay(
                X=X,
                Su=Su,
                G0=Gmax,
                D=D,
                output_length=output_length,
            )

        # parse multipliers and apply results
        t_mult = self.t_multiplier if isinstance(self.t_multiplier, float) else self.t_multiplier(X)
        m_mult = self.m_multiplier if isinstance(self.m_multiplier, float) else self.m_multiplier(X)

        return t * t_mult, m * m_mult

    def Mb_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Su_t, Su_b = from_list2x_parse_top_bottom(self.Su)
        Su = Su_t + (Su_b - Su_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        y, Mb = Mb_curves.cowden_clay(
            X=X,
            Su=Su,
            G0=Gmax,
            D=D,
            L=L,
            output_length=output_length,
        )

        return y, Mb


@dataclass(config=PydanticConfigFrozen)
class Dunkirk_sand(LateralModel):
    """A class to establish the PISA Dunkirk sand model as per  Burd et al (2020) (see [BTZA20]_)..

    Parameters
    ----------
    Dr: float or list[top_value, bottom_value]
        relative density of sand. Value to range from 0 to 100. [unit: -]
    G0: float or list[top_value, bottom_value]
        Small-strain shear modulus [unit: kPa]
    p_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for p-values
    y_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for y-values
    m_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for m-values
    t_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for t-values

    See also
    --------
    :py:func:`openpile.utils.py_curves.dunkirk_sand`, :py:func:`openpile.utils.mt_curves.dunkirk_sand`,
    :py:func:`openpile.utils.Hb_curves.dunkirk_sand`, :py:func:`openpile.utils.Mb_curves.dunkirk_sand`

    """

    #: soil friction angle [deg], if a variation in values, two values can be given.
    Dr: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: small-strain shear stiffness modulus [kPa]
    G0: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: p-multiplier
    p_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: y-multiplier
    y_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0
    #: m-multiplier
    m_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: t-multiplier
    t_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0

    # spring signature which tells that API sand only has p-y curves
    # signature if of the form [p-y:True, Hb:False, m-t:False, Mb:False]
    spring_signature = np.array([True, True, True, True], dtype=bool)

    def __str__(self):
        return f"\tDunkirk sand (PISA)\n\tDr = {var_to_str(self.Dr)}%. \n\tG0 = {round(self.G0/1000,1)} MPa"

    def py_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Dr_t, Dr_b = from_list2x_parse_top_bottom(self.Dr)
        Dr = Dr_t + (Dr_b - Dr_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        y, p = py_curves.dunkirk_sand(
            sig=sig,
            X=X,
            Dr=Dr,
            G0=Gmax,
            D=D,
            L=L,
            output_length=output_length,
        )

        # parse multipliers and apply results
        y_mult = self.y_multiplier if isinstance(self.y_multiplier, float) else self.y_multiplier(X)
        p_mult = self.p_multiplier if isinstance(self.p_multiplier, float) else self.p_multiplier(X)

        return y * y_mult, p * p_mult

    def Hb_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Dr_t, Dr_b = from_list2x_parse_top_bottom(self.Dr)
        Dr = Dr_t + (Dr_b - Dr_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        y, Hb = Hb_curves.dunkirk_sand(
            sig=sig,
            X=X,
            Dr=Dr,
            G0=Gmax,
            D=D,
            L=L,
            output_length=output_length,
        )

        return y, Hb

    def mt_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Dr_t, Dr_b = from_list2x_parse_top_bottom(self.Dr)
        Dr = Dr_t + (Dr_b - Dr_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        _, p_array = py_curves.dunkirk_sand(
            sig=sig,
            X=X,
            Dr=Dr,
            G0=Gmax,
            D=D,
            L=L,
            output_length=output_length,
        )

        m = np.zeros((output_length, output_length), dtype=np.float32)
        t = np.zeros((output_length, output_length), dtype=np.float32)

        for count, p_iter in enumerate(p_array):
            t[count, :], m[count] = mt_curves.dunkirk_sand(
                sig=sig,
                X=X,
                Dr=Dr,
                G0=Gmax,
                p=p_iter,
                D=D,
                L=L,
                output_length=output_length,
            )

        # parse multipliers and apply results
        t_mult = self.t_multiplier if isinstance(self.t_multiplier, float) else self.t_multiplier(X)
        m_mult = self.m_multiplier if isinstance(self.m_multiplier, float) else self.m_multiplier(X)

        return t * t_mult, m * m_mult

    def Mb_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Dr
        Dr_t, Dr_b = from_list2x_parse_top_bottom(self.Dr)
        Dr = Dr_t + (Dr_b - Dr_t) * depth_from_top_of_layer / layer_height
        # define G0
        G0_t, G0_b = from_list2x_parse_top_bottom(self.G0)
        Gmax = G0_t + (G0_b - G0_t) * depth_from_top_of_layer / layer_height

        y, Mb = Mb_curves.dunkirk_sand(
            sig=sig,
            X=X,
            Dr=Dr,
            G0=Gmax,
            D=D,
            L=L,
            output_length=output_length,
        )

        return y, Mb


@dataclass(config=PydanticConfigFrozen)
class API_sand(LateralModel):
    """A class to establish the API sand model.

    Parameters
    ----------
    phi: float or list[top_value, bottom_value]
        internal angle of friction in degrees
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    G0: float or list[top_value, bottom_value] or None
        Small-strain shear modulus [unit: kPa], by default None
    p_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for p-values
    y_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for y-values
    extension: str, by default None
        turn on extensions by calling them in this variable
        for API_sand, rotational springs can be added to the model with the extension "mt_curves"

    See also
    --------
    :py:func:`openpile.utils.py_curves.api_sand`

    """

    #: soil friction angle [deg], if a variation in values, two values can be given.
    phi: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: types of curves, can be of ("static","cyclic")
    kind: Literal["static", "cyclic"] = "static"
    #: small-strain stiffness [kPa], if a variation in values, two values can be given.
    G0: Optional[Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]] = None
    #: p-multiplier
    p_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: y-multiplier
    y_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0
    #: extensions available for soil model
    extension: Optional[Literal["mt_curves"]] = None

    # define class variables needed for all soil models
    m_multiplier = 1.0
    t_multiplier = 1.0

    def __post_init__(self):
        # spring signature which tells that API sand only has p-y curves in normal conditions
        # signature if e.g. of the form [p-y:True, Hb:False, m-t:False, Mb:False]
        if self.extension == "mt_curves":
            self.spring_signature = np.array([True, False, True, False], dtype=bool)
        else:
            self.spring_signature = np.array([True, False, False, False], dtype=bool)

    def __str__(self):
        return f"\tAPI sand\n\tphi = {var_to_str(self.phi)}°\n\t{self.kind} curves"

    def py_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define phi
        phi_t, phi_b = from_list2x_parse_top_bottom(self.phi)
        phi = phi_t + (phi_b - phi_t) * depth_from_top_of_layer / layer_height

        y, p = py_curves.api_sand(
            sig=sig,
            X=X,
            phi=phi,
            D=D,
            kind=self.kind,
            below_water_table=below_water_table,
            ymax=ymax,
            output_length=output_length,
        )

        # parse multipliers and apply results
        y_mult = self.y_multiplier if isinstance(self.y_multiplier, float) else self.y_multiplier(X)
        p_mult = self.p_multiplier if isinstance(self.p_multiplier, float) else self.p_multiplier(X)

        return y * y_mult, p * p_mult

    def mt_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define phi
        phi_t, phi_b = from_list2x_parse_top_bottom(self.phi)
        phi = phi_t + (phi_b - phi_t) * depth_from_top_of_layer / layer_height

        # define p vector
        _, p = py_curves.api_sand(
            sig=sig,
            X=X,
            phi=phi,
            D=D,
            kind=self.kind,
            below_water_table=below_water_table,
            ymax=ymax,
            output_length=output_length,
        )

        if p.max() > 0:
            p_norm = p / p.max()
        else:
            p_norm = p

        m = np.zeros((output_length, output_length), dtype=np.float32)
        t = np.zeros((output_length, output_length), dtype=np.float32)

        z, tz = tz_curves.api_sand(
            sig=sig,
            delta=phi - 5,
            K=0.8,
            tensile_factor=1.0,
            output_length=output_length,
        )

        # trasnform tz vector
        tz_pos = tz[tz >= 0]
        z_pos = z[z >= 0]
        diff_length_t = output_length - len(tz_pos)
        diff_length_z = output_length - len(z_pos)
        tz_pos = np.append(tz_pos, [tz_pos[-1]] * diff_length_t)
        z_pos = np.append(z_pos, [z_pos[-1] + i * 0.1 for i in range(diff_length_z)])

        t = np.arctan(z_pos.reshape((1, -1)) / (0.5 * D)) * np.ones((output_length, 1))
        m = 1 / 4 * np.pi * D**2 * tz_pos.reshape((1, -1)) * p_norm.reshape((-1, 1))

        return t, m


@dataclass(config=PydanticConfigFrozen)
class API_clay(LateralModel):
    """A class to establish the API clay model.

    Parameters
    ----------
    Su: float or list[top_value, bottom_value]
        Undrained shear strength in kPa
    eps50: float or list[top_value, bottom_value]
        strain at 50% failure load [-]
    J: float
        empirical factor varying depending on clay stiffness, varies between 0.25 and 0.50
    stiff_clay_threshold: float
        undrained shear strength [kPa] at which stiff clay curve is computed
    G0: float or list[top_value, bottom_value] or None
        Small-strain shear modulus [unit: kPa], by default None
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    p_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for p-values
    y_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for y-values
    extension: str, by default None
        turn on extensions by calling them in this variable
        for API_clay, rotational springs can be added to the model with the extension "mt_curves"


    See also
    --------
    :py:func:`openpile.utils.py_curves.api_clay`

    """

    #: undrained shear strength [kPa], if a variation in values, two values can be given.
    Su: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: strain at 50% failure load [-], if a variation in values, two values can be given.
    eps50: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: types of curves, can be of ("static","cyclic")
    kind: Literal["static", "cyclic"] = "static"
    #: small-strain stiffness [kPa], if a variation in values, two values can be given.
    G0: Optional[Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]] = None
    #: empirical factor varying depending on clay stiffness
    J: confloat(ge=0.25, le=0.5) = 0.5
    #: undrained shear strength [kPa] at which stiff clay curve is computed
    stiff_clay_threshold: PositiveFloat = 96
    #: p-multiplier
    p_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: y-multiplier
    y_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0
    #: extensions available for soil model
    extension: Optional[Literal["mt_curves"]] = None

    # define class variables needed for all soil models
    m_multiplier = 1.0
    t_multiplier = 1.0

    def __post_init__(self):
        # spring signature which tells that API clay only has p-y curves in normal conditions
        # signature if e.g. of the form [p-y:True, Hb:False, m-t:False, Mb:False]
        if self.extension == "mt_curves":
            self.spring_signature = np.array([True, False, True, False], dtype=bool)
        else:
            self.spring_signature = np.array([True, False, False, False], dtype=bool)

    def __str__(self):
        return f"\tAPI clay\n\tSu = {var_to_str(self.Su)} kPa\n\teps50 = {var_to_str(self.eps50)}\n\t{self.kind} curves"

    def py_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Su
        Su_t, Su_b = from_list2x_parse_top_bottom(self.Su)
        Su = Su_t + (Su_b - Su_t) * depth_from_top_of_layer / layer_height

        # define eps50
        eps50_t, eps50_b = from_list2x_parse_top_bottom(self.eps50)
        eps50 = eps50_t + (eps50_b - eps50_t) * depth_from_top_of_layer / layer_height

        y, p = py_curves.api_clay(
            sig=sig,
            X=X,
            Su=Su,
            eps50=eps50,
            D=D,
            J=self.J,
            stiff_clay_threshold=self.stiff_clay_threshold,
            kind=self.kind,
            ymax=ymax,
            output_length=output_length,
        )

        # parse multipliers and apply results
        y_mult = self.y_multiplier if isinstance(self.y_multiplier, float) else self.y_multiplier(X)
        p_mult = self.p_multiplier if isinstance(self.p_multiplier, float) else self.p_multiplier(X)

        return y * y_mult, p * p_mult

    def mt_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Su
        Su_t, Su_b = from_list2x_parse_top_bottom(self.Su)
        Su = Su_t + (Su_b - Su_t) * depth_from_top_of_layer / layer_height

        m = np.zeros((output_length, output_length), dtype=np.float32)
        t = np.zeros((output_length, output_length), dtype=np.float32)

        z, tz = tz_curves.api_clay(
            sig=sig,
            Su=Su,
            D=D,
            residual=1.0,
            tensile_factor=1.0,
            output_length=output_length,
        )

        # trasnform tz vector so that we only get the positive side of the vectors
        tz_pos = tz[tz >= 0]
        z_pos = z[z >= 0]
        # check how many elements we got rid off
        diff_length_t = output_length - len(tz_pos)
        diff_length_z = output_length - len(z_pos)
        # add new elements at the end of the positive only vectors
        tz_pos = np.append(tz_pos, [tz_pos[-1]] * diff_length_t)
        z_pos = np.append(z_pos, [z_pos[-1] + i * 0.1 for i in range(diff_length_z)])

        # calculate m and t vectors (they are all the same for clay)
        t = np.arctan(z_pos.reshape((1, -1)) / (0.5 * D)) * np.ones((output_length, 1))
        m = 1 / 4 * np.pi * D**2 * tz_pos.reshape((1, -1)) * np.ones((output_length, 1))

        return t, m


@dataclass(config=PydanticConfigFrozen)
class Reese_weakrock(LateralModel):
    """A class to establish the Reese weakrock model.

    Parameters
    ----------
    Ei: float or list[top_value, bottom_value]
        Initial modulus of rock [unit: kPa]
    qu: float or list[top_value, bottom_value]
        compressive strength of rock [unit: kPa]
    RQD: float or list[top_value, bottom_value]
        Rock Quality Designation [unit: %]
    k: float
        dimensional constant randing from 0.0005 to 0.00005, by default 0.0005
    ztop: float
        absolute depth of top layer elevation with respect to rock surface [m]
    p_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for p-values
    y_multiplier: float or function taking the depth as argument and returns the multiplier
        multiplier for y-values

    """

    #: initial modulus of rock [kPa], if a variation in values, two values can be given.
    Ei: Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]
    #: scompressive strength of rock [kPa], if a variation in values, two values can be given.
    qu: Optional[Union[PositiveFloat, conlist(PositiveFloat, min_items=1, max_items=2)]]
    #: Rock Quality Designation
    RQD: confloat(ge=0.0, le=100.0)
    #: dimnesional constant
    k: confloat(ge=0.00005, le=0.0005)
    #: absolute depth of top layer elevation with respect to rock surface [m]
    ztop: confloat(ge=0.0)
    #: p-multiplier
    p_multiplier: Union[Callable[[float], float], confloat(ge=0.0)] = 1.0
    #: y-multiplier
    y_multiplier: Union[Callable[[float], float], confloat(gt=0.0)] = 1.0

    # define class variables needed for all soil models
    m_multiplier = 1.0
    t_multiplier = 1.0

    spring_signature = np.array([True, False, False, False], dtype=bool)

    def __str__(self):
        return f"\tReese weakrock\n\tEi = {var_to_str(self.Ei)}kPa\n\tqu = {var_to_str(self.qu)}kPa\n\tRQD = {var_to_str(self.RQD)}%"

    def py_spring_fct(
        self,
        sig: float,
        X: float,
        layer_height: float,
        depth_from_top_of_layer: float,
        D: float,
        L: float = None,
        below_water_table: bool = True,
        ymax: float = 0.0,
        output_length: int = 15,
    ):
        # validation
        if depth_from_top_of_layer > layer_height:
            raise ValueError("Spring elevation outside layer")

        # define Ei
        Ei_t, Ei_b = from_list2x_parse_top_bottom(self.Ei)
        Ei = Ei_t + (Ei_b - Ei_t) * depth_from_top_of_layer / layer_height

        # define qu
        qu_t, qu_b = from_list2x_parse_top_bottom(self.qu)
        qu = qu_t + (qu_b - qu_t) * depth_from_top_of_layer / layer_height

        y, p = py_curves.reese_weakrock(
            Ei=Ei,
            xr=(X + self.ztop),
            RQD=self.RQD,
            qu=qu,
            D=D,
            k=self.k,
            output_length=output_length,
        )

        # parse multipliers and apply results
        y_mult = self.y_multiplier if isinstance(self.y_multiplier, float) else self.y_multiplier(X)
        p_mult = self.p_multiplier if isinstance(self.p_multiplier, float) else self.p_multiplier(X)

        return y * y_mult, p * p_mult



"""
`Construct` module
==================

The `construct` module is used to construct all objects that 
form the inputs to calculations in openpile. 


These objects include:

- the Pile
- the SoilProfile
  - the Layer
- the Model

**Usage**

>>> from openpile.construct import Pile, SoilProfile, Layer, Model

"""

import math as m
import pandas as pd
import numpy as np
import warnings
from typing import List, Dict, Optional, Union
from typing_extensions import Literal
from pydantic import (
    BaseModel,
    Field,
    root_validator,
    validator,
    PositiveFloat,
    confloat,
    conlist,
    constr,
    Extra,
    ValidationError,
)
from pydantic.dataclasses import dataclass
import matplotlib.pyplot as plt

import openpile.utils.graphics as graphics
import openpile.core.validation as validation
import openpile.soilmodels as soilmodels

from openpile.core import misc

from openpile.soilmodels import ConstitutiveModel

from openpile.core.misc import generate_color_string


class PydanticConfig:
    arbitrary_types_allowed = True
    extra = Extra.forbid
    post_init_call = "after_validation"


@dataclass(config=PydanticConfig)
class Pile:
    """
    A class to create the pile.

    Parameters
    ----------
    name : str
        Pile/Structure's name.
    top_elevation : float
        top elevation of the pile. Note that this elevation provides a reference point to
        know where the pile is located, especially with respect to other object such as a SoilProfile.
    pile_sections : Dict[str, List[float]]
        argument that stores the relevant data of each pile segment.
        Below are the needed keys for the available piles:
        - kind:'Circular' >> keys:['length', 'diameter', 'wall thickness']
    kind : Literal["Circular",]
        type of pile or type of cross-section. by default "Circular"
    material : Literal["Steel",]
        material the pile is made of. by default "Steel"


    Example
    -------

    >>> from openpile.construct import Pile

    >>> # Create a pile instance with two sections of respectively 10m and 30m length.
    >>> pile = Pile(name = "",
    >>>         kind='Circular',
    >>>         material='Steel',
    >>>         top_elevation = 0,
    >>>         pile_sections={
    >>>             'length':[10,30],
    >>>             'diameter':[7.5,7.5],
    >>>             'wall thickness':[0.07, 0.08],
    >>>         }
    >>>     )
    """

    #: name of the pile
    name: str
    #: select the type of pile, can be of ('Circular', )
    kind: Literal["Circular"]
    #: select the type of material the pile is made of, can be of ('Steel', )
    material: Literal["Steel"]
    #: top elevation of the pile according to general vertical reference set by user
    top_elevation: float
    #: pile geometry made of a dictionary of lists. the structure of the dictionary depends on the type of pile selected.
    #: There can be as many sections as needed by the user. The length of the listsdictates the number of pile sections.
    pile_sections: Dict[str, List[PositiveFloat]]

    def __post_init__(self):
        # check that dict is correctly entered
        validation.pile_sections_must_be(self)

        # Create material specific specs for given material
        # if steel
        if self.material == "Steel":
            # unit weight
            self._uw = 78.0  # kN/m3
            # young modulus
            self._young_modulus = 210.0e6  # kPa
            # Poisson's ratio
            self._nu = 0.3
        else:
            raise UserWarning

        self._shear_modulus = self._young_modulus / (2 + 2 * self._nu)

        # create pile data used by openpile for mesh and calculations.
        # Create top and bottom elevations
        elevation = []
        # add bottom of section i and top of section i+1 (essentially the same values)
        for idx, val in enumerate(self.pile_sections["length"]):
            if idx == 0:
                elevation.append(self.top_elevation)
                elevation.append(elevation[-1] - val)
            else:
                elevation.append(elevation[-1])
                elevation.append(elevation[-1] - val)

        # create sectional properties

        # spread
        diameter = []
        # add top and bottom of section i (essentially the same values)
        for idx, val in enumerate(self.pile_sections["diameter"]):
            diameter.append(val)
            diameter.append(diameter[-1])

        # thickness
        thickness = []
        # add top and bottom of section i (essentially the same values)
        for idx, val in enumerate(self.pile_sections["wall thickness"]):
            thickness.append(val)
            thickness.append(thickness[-1])

        # Area & second moment of area
        area = []
        second_moment_of_area = []
        # add top and bottom of section i (essentially the same values)
        for _, (d, wt) in enumerate(
            zip(self.pile_sections["diameter"], self.pile_sections["wall thickness"])
        ):
            # calculate area
            if self.kind == "Circular":
                A = m.pi / 4 * (d**2 - (d - 2 * wt) ** 2)
                I = m.pi / 64 * (d**4 - (d - 2 * wt) ** 4)
                area.append(A)
                area.append(area[-1])
                second_moment_of_area.append(I)
                second_moment_of_area.append(second_moment_of_area[-1])
            else:
                # not yet supporting other kind
                raise ValueError()

        # Create pile data
        self.data = pd.DataFrame(
            data={
                "Elevation [m]": elevation,
                "Diameter [m]": diameter,
                "Wall thickness [m]": thickness,
                "Area [m2]": area,
                "I [m4]": second_moment_of_area,
            }
        )

    def __str__(self):
        return self.data.to_string()

    @property
    def bottom_elevation(self) -> float:
        """
        Bottom elevation of the pile [m VREF].
        """
        return self.top_elevation - sum(self.pile_sections["length"])

    @property
    def length(self) -> float:
        """
        Pile length [m].
        """
        return sum(self.pile_sections["length"])

    @property
    def volume(self) -> float:
        """
        Pile volume [m3].
        """
        A = self.data["Area [m2]"].values[1:]
        L = np.abs(np.diff(self.data["Elevation [m]"].values))
        return round((A * L).sum(), 2)

    @property
    def weight(self) -> float:
        """
        Pile weight [kN].
        """
        return round(self.volume * self._uw, 2)

    @property
    def E(self) -> float:
        """
        Young modulus of the pile material [kPa]. Thie value does not vary across and along the pile.
        """
        try:
            return self._young_modulus
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @E.setter
    def E(self, value: float) -> None:
        try:
            self._young_modulus = value
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @property
    def I(self) -> float:
        """
        Second moment of area of the pile [m4].

        The user can use the method :py:meth:`openpile.construct.Pile.set_I` to customise the second
        moment of area for different sections of the pile.
        """
        try:
            return self.data["I [m4]"]
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @property
    def width(self) -> float:
        """
        Width of the pile [m]. (Used to compute soil springs)
        """
        try:
            return self.data.loc[:, "Diameter [m]"]
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @width.setter
    def width(self, value: float) -> None:
        try:
            self.data.loc[:, "Diameter [m]"] = value
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @property
    def area(self) -> float:
        "Sectional area of the pile [m2]"
        try:
            return self.data.loc[:, "Area [m2]"]
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @area.setter
    def area(self, value: float) -> None:
        try:
            self.data.loc[:, "Area [m2]"] = value
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            print(e)

    @classmethod
    def create(
        cls,
        name: str,
        top_elevation: float,
        pile_sections: Dict[str, List[float]],
        kind: Literal[
            "Circular",
        ] = "Circular",
        material: Literal[
            "Steel",
        ] = "Steel",
    ):
        """A method to create the pile.

        Parameters
        ----------
        name : str
            Pile/Structure's name.
        top_elevation : float
            top elevation of the pile. Note that this elevation provides a reference point to
            know where the pile is located, especially with respect to other object such as a SoilProfile.
        pile_sections : Dict[str, List[float]]
            argument that stores the relevant data of each pile segment.
            Below are the needed keys for the available piles:
            - kind:'Circular' >> keys:['length', 'diameter', 'wall thickness']
        kind : Literal["Circular",]
            type of pile or type of cross-section. by default "Circular"
        material : Literal["Steel",]
            material the pile is made of. by default "Steel"

        Returns
        -------
        openpile.construct.Pile
            a Pile instance with embedded postprocessing to perform calculations with openpile.
        """

        obj = cls(
            name=name,
            kind=kind,
            material=material,
            top_elevation=top_elevation,
            pile_sections=pile_sections,
        )

        warnings.warn(
            "\nThe method Pile.create() will be removed in version 1.0.0."
            "\nPlease use the base class to create a Pile instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return obj

    @classmethod
    def create_tubular(
        cls,
        name: str,
        top_elevation: float,
        bottom_elevation: float,
        diameter: float,
        wt: float,
        material: str = "Steel",
    ):
        """A method to simplify the creation of a Pile instance.
        This method creates a circular and hollow pile of constant diameter and wall thickness.

        Parameters
        ----------
        name : str
            Pile/Structure's name.
        top_elevation : float
            top elevation of the pile [m VREF]
        bottom_elevation : float
            bottom elevation of the pile [m VREF]
        diameter : float
            pile diameter [m]
        wt : float
            pile's wall thickness [m]
        material : Literal["Steel",]
            material the pile is made of. by default "Steel"

        Returns
        -------
        openpile.construct.Pile
            a Pile instance.
        """

        obj = cls(
            name=name,
            kind="Circular",
            material=material,
            top_elevation=top_elevation,
            pile_sections={
                "length": [
                    (top_elevation - bottom_elevation),
                ],
                "wall thickness": [
                    wt,
                ],
                "diameter": [
                    diameter,
                ],
            },
        )

        return obj

    def set_I(self, value: float, section: int) -> None:
        """set second moment of area for a particular section of the pile.

        Parameters
        ----------
        value : float
            new second moment of area [m4].
        section : int
            section number for which to set new second moment of area

        """
        try:
            length = len(self.data["I [m4]"].values)
            if section * 2 > length:
                print("section number is too large")
            elif section < 1:
                print("section number must be 1 or above")
            else:
                self.data.loc[section * 2 - 2, "I [m4]"] = value
                self.data.loc[section * 2 - 1, "I [m4]"] = value
        except AttributeError:
            print("Please first create the pile with the Pile.create() method")
        except Exception as e:
            raise Exception

    def plot(self, assign=False):
        """Creates a plot of the pile with the properties.

        Parameters
        ----------
        assign : bool, optional
            this parameter can be set to True to return the figure, by default False

        Returns
        -------
        matplotlib.pyplot.figure
            only return the object if assign=True

        Example
        -------

        .. image:: _static/plots/Pile_plot.png
           :width: 70%

        """
        fig = graphics.pile_plot(self)
        return fig if assign else None


@dataclass(config=PydanticConfig)
class Layer:
    """A class to create a layer.

    The Layer stores information on the soil parameters of the layer as well
    as the relevant/representative constitutive model (aka. the soil spring).

    Parameters
    ----------
    name : str
        Name of the layer, use for printout.
    top : float
        top elevation of the layer in [m].
    bottom : float
        bottom elevation of the layer in [m].
    weight : float
        total unit weight in [kN/m3], cannot be lower than 10.
    lateral_model : ConstitutiveModel
        Lateral soil model of the layer, by default None.
    axial_model : ConstitutiveModel
        Axial soil model of the layer, by default None.
    color : str
        soil layer color in HEX format (e.g. '#000000'), by default None.
        If None, the color is generated randomly.


    Example
    -------

    >>> from openpile.construct import Layer
    >>> from openpile.core.soilmodels import API_clay

    >>> # Create a layer with increasing values of Su and eps50
    >>> layer1 = Layer(name='Soft Clay',
                   top=0,
                   bottom=-10,
                   weight=19,
                   lateral_model=API_clay(Su=[30,35], eps50=[0.01, 0.02], Neq=100),
                   )

    >>> # show layer
    >>> print(layer1)
    Name: Soft Clay
    Elevation: (0.0) - (-10.0) m
    Weight: 19.0 kN/m3
    Lateral model: 	API clay
        Su = 30.0-35.0 kPa
        eps50 = 0.01-0.02
        Cyclic, N = 100 cycles
    Axial model: None
    """

    #: name of the layer, use for printout
    name: str
    #: top elevaiton of the layer
    top: float
    #: bottom elevaiton of the layer
    bottom: float
    #: unit weight in kN of the layer
    weight: confloat(gt=10.0)
    #: Lateral constitutive model of the layer
    lateral_model: Optional[ConstitutiveModel] = None
    #: Axial constitutive model of the layer
    axial_model: Optional[ConstitutiveModel] = None
    #: Layer's color when plotted
    color: Optional[constr(min_length=7, max_length=7)] = None

    def __post_init__(self):
        if self.color is None:
            self.color = generate_color_string("earth")

    def __str__(self):
        return f"Name: {self.name}\nElevation: ({self.top}) - ({self.bottom}) m\nWeight: {self.weight} kN/m3\nLateral model: {self.lateral_model}\nAxial model: {self.axial_model}"

    @root_validator
    def check_elevations(cls, values):  # pylint: disable=no-self-argument
        if not values["top"] > values["bottom"]:
            print("Bottom elevation is higher than top elevation")
            raise ValueError
        else:
            return values


@dataclass(config=PydanticConfig)
class SoilProfile:
    """
    A class to create the soil profile. A soil profile consist of a ground elevation (or top elevation)
    with one or more layers of soil.

    Additionally, a soil profile can include discrete information at given elevation such as CPT
    (Cone Penetration Test) data. Not Implemented yet!

    Parameters
    ----------
    name : str
        Name of the soil profile, used for printout and plots.
    top_elevation : float
        top elevation of the soil profile in [m VREF].
    water_line : float
        elevation of the water table in [m VREF].
    layers : list[Layer]
        list of layers for the soil profile.
    cpt_data : np.ndarray
        cpt data table with
        1st col: elevation [m],
        2nd col: cone resistance [kPa],
        3rd col: sleeve friction [kPa],
        4th col: pore pressure u2 [kPa].

    Example
    -------

    >>> from openpile.construct import SoilProfile, Layer
    >>> from openpile.core.soilmodels import API_sand, API_clay

    >>> # Create a two-layer soil profile
    >>> sp = SoilProfile(
    >>>     name="BH01",
    >>>     top_elevation=0,
    >>>     water_line=0,
    >>>     layers=[
    >>>         Layer(
    >>>             name='Layer0',
    >>>             top=0,
    >>>             bottom=-20,
    >>>             weight=18,
    >>>             lateral_model= API_sand(phi=30, Neq=100)
    >>>         ),
    >>>         Layer( name='Layer1',
    >>>                 top=-20,
    >>>                 bottom=-40,
    >>>                 weight=19,
    >>>                 lateral_model= API_clay(Su=50, eps50=0.01, Neq=100),)
    >>>     ]
    >>> )

    >>> # Check soil profile content
    >>> print(sp)
    Layer 1
    ------------------------------
    Name: Layer0
    Elevation: (0.0) - (-20.0) m
    Weight: 18.0 kN/m3
    Lateral model: 	API sand
        phi = 30.0°
        Cyclic, N = 100 cycles
    Axial model: None
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
    Layer 2
    ------------------------------
    Name: Layer1
    Elevation: (-20.0) - (-40.0) m
    Weight: 19.0 kN/m3
    Lateral model: 	API clay
        Su = 50.0 kPa
        eps50 = 0.01
        Cyclic, N = 100 cycles
    Axial model: None
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    """

    #: name of soil profile / borehole / location
    name: str
    #: top of ground elevation with respect to the model reference elevation datum
    top_elevation: float
    #: water elevation (this can refer to sea elevation of water table)
    water_line: float
    #: soil layers to consider in the soil propfile
    layers: List[Layer]
    #: Cone Penetration Test data with folloeing structure:
    #: 1st col: elevation[m],
    #: 2nd col: cone resistance[kPa],
    #: 3rd col: sleeve friction [kPa]
    #: 4th col: pore pressure u2 [kPa]
    #: (the cpt data outside the soil profile boundaries will be ignored)
    cpt_data: Optional[np.ndarray] = None

    @root_validator
    def check_layers_elevations(cls, values):  # pylint: disable=no-self-argument
        layers = values["layers"]

        top_elevations = np.array([x.top for x in layers], dtype=float)
        bottom_elevations = np.array([x.bottom for x in layers], dtype=float)
        idx_sort = np.argsort(top_elevations)

        top_sorted = top_elevations[idx_sort][::-1]
        bottom_sorted = bottom_elevations[idx_sort][::-1]

        # check no overlap
        if top_sorted[0] != values["top_elevation"]:
            raise ValueError("top_elevation not matching uppermost layer's elevations.")

        for i in range(len(top_sorted) - 1):
            if not m.isclose(top_sorted[i + 1], bottom_sorted[i], abs_tol=0.001):
                raise ValueError("Layers' elevations overlap.")

        return values

    @root_validator
    def check_multipliers_in_lateral_model(cls, values):
        def check_multipliers_callable(multiplier, ground_level, top, bottom, type):
            # if not a float, it must be a callable, then we check for Real Positive float
            if not isinstance(multiplier, float):
                # defines depth below ground to check
                depths = ground_level - np.linspace(start=top, stop=bottom, num=100)
                # check if positive real float is returned
                for depth in depths:
                    result = multiplier(depth)
                    if not isinstance(result, float):
                        TypeError(
                            f"One or more results of the {type}-multiplier callable is not a float"
                        )
                        return None
                    else:
                        if type in ["p", "m"]:
                            if result < 0.0:
                                print(
                                    f"One or more results of the {type}-multiplier callable is negative"
                                )
                                return None
                        elif type in ["y", "t"]:
                            if not result > 0.0:
                                ValueError(
                                    f"One or more results of the {type}-multiplier callable is not strictly positive"
                                )
                                return None

        layers = values["layers"]

        for layer in layers:
            if layer.lateral_model is not None:
                # check p-multipliers
                check_multipliers_callable(
                    layer.lateral_model.p_multiplier,
                    values["top_elevation"],
                    layer.top,
                    layer.bottom,
                    "p",
                )
                # check y-multipliers
                check_multipliers_callable(
                    layer.lateral_model.y_multiplier,
                    values["top_elevation"],
                    layer.top,
                    layer.bottom,
                    "y",
                )
                # check m-multipliers
                check_multipliers_callable(
                    layer.lateral_model.m_multiplier,
                    values["top_elevation"],
                    layer.top,
                    layer.bottom,
                    "m",
                )
                # check t-multipliers
                check_multipliers_callable(
                    layer.lateral_model.t_multiplier,
                    values["top_elevation"],
                    layer.top,
                    layer.bottom,
                    "t",
                )

        return values

    def __post_init__(self):
        pass

    def __str__(self):
        """List all layers in table-like format"""
        out = ""
        i = 0
        for layer in self.layers:
            i += 1
            out += f"Layer {i}\n" + "-" * 30 + "\n"
            out += f"{layer}\n" + "~" * 30 + "\n"
        return out

    @property
    def bottom_elevation(self) -> float:
        """
        Bottom elevation of the soil profile [m VREF].
        """
        return self.top_elevation - sum([abs(x.top - x.bottom) for x in self.layers])

    def plot(self, assign=False):
        """Creates a plot illustrating the stratigraphy.

        Parameters
        ----------
        assign : bool, optional
            this parameter can be set to True to return the figure, by default False

        Returns
        -------
        matplotlib.pyplot.figure
            only return the object if assign=True

        Example
        -------

        .. image:: _static/plots/SoilProfile_plot.png
           :scale: 70%
        """
        fig = graphics.soil_plot(self)
        return fig if assign is True else None


@dataclass(config=PydanticConfig)
class Model:
    """
    A class to create a Model.

    A Model is constructed based on the pile geometry/data primarily.
    Additionally, a soil profile can be fed to the Model, and soil springs can be created.

    Parameters
    ----------
    name : str
        Name of the model
    pile : Pile
        Pile instance to be included in the model.
    soil : Optional[SoilProfile], optional
        SoilProfile instance, by default None.
    element_type : str, optional
        can be of ['EulerBernoulli','Timoshenko'], by default 'Timoshenko'.
    x2mesh : List[float], optional
        additional elevations to be included in the mesh, by default none.
    coarseness : float, optional
        maximum distance in meters between two nodes of the mesh, by default 0.5.
    distributed_lateral : bool, optional
        include distributed lateral springs, by default True.
    distributed_moment : bool, optional
        include distributed moment springs, by default False.
    base_shear : bool, optional
        include lateral spring at pile toe, by default False.
    base_moment : bool, optional
        include moment spring at pile toe, by default False.


    Example
    -------

    >>> from openpile.construct import Pile, Model, Layer
    >>> from openpile.core.soilmodels import API_sand

    >>> # create pile
    >>> p = Pile(name = "WTG01",
    >>> 		kind='Circular',
    >>> 		material='Steel',
    >>> 		top_elevation = 0,
    >>> 		pile_sections={
    >>> 			'length':[10,30],
    >>> 			'diameter':[7.5,7.5],
    >>> 			'wall thickness':[0.07, 0.08],
    >>> 		}
    >>> 	)

    >>> # Create Soil Profile
    >>> sp = SoilProfile(
    >>> 	name="BH01",
    >>> 	top_elevation=0,
    >>> 	water_line=0,
    >>> 	layers=[
    >>> 		Layer(
    >>> 			name='Layer0',
    >>> 			top=0,
    >>> 			bottom=-40,
    >>> 			weight=18,
    >>> 			lateral_model= API_sand(phi=30, Neq=100)
    >>> 		),
    >>> 	]
    >>> )

    >>> # Create Model
    >>> M = Model(name="Example", pile=p, soil=sp)

    >>> # create Model without soil maximum 5 metres apart.
    >>> Model_without_soil = Model(name = "Example without soil", pile=p, coarseness=5)
    >>> # create Model with nodes maximum 1 metre apart with soil profile
    >>> Model_with_soil = Model(name = "Example with soil", pile=p, soil=sp, coarseness=1)

    """

    #: model name
    name: str
    #: pile instance that the Model should consider
    pile: Pile
    #: soil profile instance that the Model should consider
    soil: Optional[SoilProfile] = None
    #: type of beam elements
    element_type: Literal["Timoshenko", "EulerBernoulli"] = "Timoshenko"
    #: x coordinates values to mesh as nodes
    x2mesh: List[float] = Field(default_factory=list)
    #: mesh coarseness, represent the maximum accepted length of elements
    coarseness: float = 0.5
    #: whether to include p-y springs in the calculations
    distributed_lateral: bool = True
    #: whether to include m-t springs in the calculations
    distributed_moment: bool = True
    #: whether to include Hb-y spring in the calculations
    base_shear: bool = True
    #: whether to include Mb-t spring in the calculations
    base_moment: bool = True
    #: whether to include t-z springs in the calculations
    distributed_axial: bool = False
    #: whether to include Q-z spring in the calculations
    base_axial: bool = False

    @root_validator(skip_on_failure=True)
    def soil_and_pile_bottom_elevation_match(cls, values):  # pylint: disable=no-self-argument
        if values["soil"] is None:
            pass
        else:
            if values["pile"].bottom_elevation < values["soil"].bottom_elevation:
                raise UserWarning("The pile ends deeper than the soil profile.")
        return values

    def __post_init__(self):
        def check_springs(arr):
            check_nan = np.isnan(arr).any()
            check_negative = (arr < 0).any()

            return check_nan or check_negative

        def get_coordinates() -> pd.DataFrame:
            # Primary discretisation over x-axis
            x = np.array([], dtype=np.float16)
            # add get pile relevant sections
            x = np.append(x, self.pile.data["Elevation [m]"].values)
            # add soil relevant layers and others
            if self.soil is not None:
                soil_elevations = np.array(
                    [x.top for x in self.soil.layers] + [x.bottom for x in self.soil.layers],
                    dtype=float,
                )
                if any(soil_elevations < self.pile.bottom_elevation):
                    soil_elevations = np.append(self.pile.bottom_elevation, soil_elevations)
                    soil_elevations = soil_elevations[soil_elevations >= self.pile.bottom_elevation]
                x = np.append(x, soil_elevations)
            # add user-defined elevation
            x = np.append(x, self.x2mesh)

            # get unique values and sort in reverse order
            x = np.unique(x)[::-1]

            # Secondary discretisation over x-axis depending on coarseness factor
            x_secondary = np.array([], dtype=np.float16)
            for i in range(len(x) - 1):
                spacing = x[i] - x[i + 1]
                new_spacing = spacing
                divider = 1
                while new_spacing > self.coarseness:
                    divider += 1
                    new_spacing = spacing / divider
                new_x = x[i] - (
                    np.arange(start=1, stop=divider) * np.tile(new_spacing, (divider - 1))
                )
                x_secondary = np.append(x_secondary, new_x)

            # assemble x- coordinates
            x = np.append(x, x_secondary)
            x = np.unique(x)[::-1]

            # dummy y- coordinates
            y = np.zeros(shape=x.shape)

            # create dataframe coordinates
            nodes = pd.DataFrame(
                data={
                    "x [m]": x,
                    "y [m]": y,
                },
                dtype=float,
            ).round(3)
            nodes.index.name = "Node no."

            element = pd.DataFrame(
                data={
                    "x_top [m]": x[:-1],
                    "x_bottom [m]": x[1:],
                    "y_top [m]": y[:-1],
                    "y_bottom [m]": y[1:],
                },
                dtype=float,
            ).round(3)
            element.index.name = "Element no."

            return nodes, element

            # function doing the work

        def get_soil_profile() -> pd.DataFrame:
            top_elevations = [x.top for x in self.soil.layers]
            bottom_elevations = [x.bottom for x in self.soil.layers]
            soil_weights = [x.weight for x in self.soil.layers]

            idx_sort = np.argsort(top_elevations)[::-1]

            top_elevations = [top_elevations[i] for i in idx_sort]
            soil_weights = [soil_weights[i] for i in idx_sort]
            bottom_elevations = [bottom_elevations[i] for i in idx_sort]

            # #calculate vertical stress
            # v_stress = [0.0,]
            # for uw, top, bottom in zip(soil_weights, top_elevations, bottom_elevations):
            #     v_stress.append(v_stress[-1] + uw*(top-bottom))

            # elevation in model w.r.t to x axis
            x = top_elevations

            return pd.DataFrame(
                data={"Top soil layer [m]": x, "Unit Weight [kN/m3]": soil_weights},
                dtype=np.float64,
            )

        def create_springs() -> np.ndarray:
            # dim of springs
            spring_dim = 15

            # Allocate array
            py = np.zeros(shape=(self.element_number, 2, 2, spring_dim), dtype=np.float32)
            mt = np.zeros(
                shape=(self.element_number, 2, 2, spring_dim, spring_dim), dtype=np.float32
            )
            Hb = np.zeros(shape=(1, 1, 2, spring_dim), dtype=np.float32)
            Mb = np.zeros(shape=(1, 1, 2, spring_dim), dtype=np.float32)

            tz = np.zeros(shape=(self.element_number, 2, 2, 15), dtype=np.float32)

            # fill in spring for each element
            for layer in self.soil.layers:
                elements_for_layer = self.soil_properties.loc[
                    (self.soil_properties["x_top [m]"] <= layer.top)
                    & (self.soil_properties["x_bottom [m]"] >= layer.bottom)
                ].index

                # py curve
                if layer.lateral_model is None:
                    pass
                else:
                    # Set local layer parameters for each element of the layer
                    for i in elements_for_layer:
                        # vertical effective stress
                        sig_v = self.soil_properties[
                            ["sigma_v top [kPa]", "sigma_v bottom [kPa]"]
                        ].iloc[i]
                        # elevation
                        elevation = self.soil_properties[["x_top [m]", "x_bottom [m]"]].iloc[i]
                        # depth from ground
                        depth_from_ground = (
                            self.soil_properties[["xg_top [m]", "xg_bottom [m]"]].iloc[i]
                        ).abs()
                        # pile width
                        pile_width = self.element_properties["Diameter [m]"].iloc[i]

                        # p-y curves
                        if (
                            layer.lateral_model.spring_signature[0] and self.distributed_lateral
                        ):  # True if py spring function exist

                            # calculate springs (top and) for each element
                            for j in [0, 1]:
                                (py[i, j, 1], py[i, j, 0]) = layer.lateral_model.py_spring_fct(
                                    sig=sig_v[j],
                                    X=depth_from_ground[j],
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(layer.top - elevation[j]),
                                    D=pile_width,
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=elevation[j] <= self.soil.water_line,
                                    output_length=spring_dim,
                                )

                        if (
                            layer.lateral_model.spring_signature[2] and self.distributed_moment
                        ):  # True if mt spring function exist

                            # calculate springs (top and) for each element
                            for j in [0, 1]:
                                (mt[i, j, 1], mt[i, j, 0]) = layer.lateral_model.mt_spring_fct(
                                    sig=sig_v[j],
                                    X=depth_from_ground[j],
                                    layer_height=(layer.top - layer.bottom),
                                    depth_from_top_of_layer=(layer.top - elevation[j]),
                                    D=pile_width,
                                    L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                    below_water_table=elevation[j] <= self.soil.water_line,
                                    output_length=spring_dim,
                                )

                    # check if pile tip is within layer
                    if (
                        layer.top >= self.pile.bottom_elevation
                        and layer.bottom <= self.pile.bottom_elevation
                    ):

                        # Hb curve
                        sig_v_tip = self.soil_properties["sigma_v bottom [kPa]"].iloc[-1]

                        if layer.lateral_model.spring_signature[1] and self.base_shear:

                            # calculate Hb spring
                            (Hb[0, 0, 1], Hb[0, 0, 0]) = layer.lateral_model.Hb_spring_fct(
                                sig=sig_v_tip,
                                X=-self.pile.bottom_elevation,
                                layer_height=(layer.top - layer.bottom),
                                depth_from_top_of_layer=(layer.top - self.pile.bottom_elevation),
                                D=pile_width,
                                L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                below_water_table=self.pile.bottom_elevation
                                <= self.soil.water_line,
                                output_length=spring_dim,
                            )

                        # Mb curve
                        if layer.lateral_model.spring_signature[3] and self.base_moment:

                            (Mb[0, 0, 1], Mb[0, 0, 0]) = layer.lateral_model.Mb_spring_fct(
                                sig=sig_v_tip,
                                X=-self.pile.bottom_elevation,
                                layer_height=(layer.top - layer.bottom),
                                depth_from_top_of_layer=(layer.top - self.pile.bottom_elevation),
                                D=pile_width,
                                L=(self.soil.top_elevation - self.pile.bottom_elevation),
                                below_water_table=self.pile.bottom_elevation
                                <= self.soil.water_line,
                                output_length=spring_dim,
                            )

            if check_springs(py):
                print("py springs have negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            if check_springs(mt):
                print("mt springs have negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            if check_springs(Hb):
                print("Hb spring has negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            if check_springs(Mb):
                print("Mb spring has negative or NaN values.")
                print(
                    """if using PISA type springs, this can be due to parameters behind out of the parameter space.
                Please check that: 2 < L/D < 6.
                """
                )
            return py, mt, Hb, Mb, tz

        # creates mesh coordinates
        self.nodes_coordinates, self.element_coordinates = get_coordinates()
        self.element_number = int(self.element_coordinates.shape[0])

        # creates element structural properties
        # merge Pile.data and self.coordinates
        self.element_properties = pd.merge_asof(
            left=self.element_coordinates.sort_values(by=["x_top [m]"]),
            right=self.pile.data.sort_values(by=["Elevation [m]"]),
            left_on="x_top [m]",
            right_on="Elevation [m]",
            direction="forward",
        ).sort_values(by=["x_top [m]"], ascending=False)
        # add young modulus to data
        self.element_properties["E [kPa]"] = self.pile.E
        # delete Elevation [m] column
        self.element_properties.drop("Elevation [m]", inplace=True, axis=1)
        # reset index
        self.element_properties.reset_index(inplace=True, drop=True)

        # create soil properties
        if self.soil is not None:
            self.soil_properties = pd.merge_asof(
                left=self.element_coordinates[["x_top [m]", "x_bottom [m]"]].sort_values(
                    by=["x_top [m]"]
                ),
                right=get_soil_profile().sort_values(by=["Top soil layer [m]"]),
                left_on="x_top [m]",
                right_on="Top soil layer [m]",
                direction="forward",
            ).sort_values(by=["x_top [m]"], ascending=False)
            # add elevation of element w.r.t. ground level
            self.soil_properties["xg_top [m]"] = (
                self.soil_properties["x_top [m]"] - self.soil.top_elevation
            )
            self.soil_properties["xg_bottom [m]"] = (
                self.soil_properties["x_bottom [m]"] - self.soil.top_elevation
            )
            # add vertical stress at top and bottom of each element
            condition_below_water_table = self.soil_properties["x_top [m]"] <= self.soil.water_line
            self.soil_properties["Unit Weight [kN/m3]"][condition_below_water_table] = (
                self.soil_properties["Unit Weight [kN/m3]"][condition_below_water_table] - 10.0
            )
            s = (
                self.soil_properties["x_top [m]"] - self.soil_properties["x_bottom [m]"]
            ) * self.soil_properties["Unit Weight [kN/m3]"]
            self.soil_properties["sigma_v top [kPa]"] = np.insert(
                s.cumsum().values[:-1],
                np.where(self.soil_properties["x_top [m]"].values == self.soil.top_elevation)[0],
                0.0,
            )
            self.soil_properties["sigma_v bottom [kPa]"] = s.cumsum()
            # reset index
            self.soil_properties.reset_index(inplace=True, drop=True)

            # Create arrays of springs
            (
                self._py_springs,
                self._mt_springs,
                self._Hb_spring,
                self._Mb_spring,
                self._tz_springs,
            ) = create_springs()

        # Initialise nodal global forces with link to nodes_coordinates (used for force-driven calcs)
        self.global_forces = self.nodes_coordinates.copy()
        self.global_forces["Px [kN]"] = 0
        self.global_forces["Py [kN]"] = 0
        self.global_forces["Mz [kNm]"] = 0

        # Initialise nodal global displacement with link to nodes_coordinates (used for displacement-driven calcs)
        self.global_disp = self.nodes_coordinates.copy()
        self.global_disp["Tx [m]"] = 0
        self.global_disp["Ty [m]"] = 0
        self.global_disp["Rz [rad]"] = 0

        # Initialise nodal global support with link to nodes_coordinates (used for defining boundary conditions)
        self.global_restrained = self.nodes_coordinates.copy()
        self.global_restrained["Tx"] = False
        self.global_restrained["Ty"] = False
        self.global_restrained["Rz"] = False

    @property
    def embedment(self) -> float:
        """Pile embedment length [m].

        Returns
        -------
        float (or None if no SoilProfile is present)
            Pile embedment
        """
        if self.soil is None:
            return None
        else:
            return self.soil.top_elevation - self.pile.bottom_elevation

    @property
    def top(self) -> float:
        """top elevation of the model [m].

        Returns
        -------
        float
        """
        if self.soil is None:
            return self.pile.top_elevation
        else:
            return max(self.pile.top_elevation, self.soil.top_elevation, self.soil.water_line)

    @property
    def bottom(self) -> float:
        """bottom elevation of the model [m].

        Returns
        -------
        float
        """

        if self.soil is None:
            return self.pile.bottom_elevation
        else:
            return min(self.pile.bottom_elevation, self.soil.bottom_elevation)

    def get_structural_properties(self) -> pd.DataFrame:
        """
        Returns a table with the structural properties of the pile sections.
        """
        try:
            return self.element_properties
        except AttributeError:
            print("Data not found. Please create Model with the Model.create() method.")
        except Exception as e:
            print(e)

    def get_soil_properties(self) -> pd.DataFrame:
        """
        Returns a table with the soil main properties and soil models of each element.
        """
        try:
            return self.soil_properties
        except AttributeError:
            print("Data not found. Please create Model with the Model.create() method.")
        except Exception as e:
            print(e)

    def get_pointload(self, output=False, verbose=True):
        """
        Returns the point loads currently defined in the mesh via printout statements.

        Parameters
        ----------
        output : bool, optional
            If true, it returns the printout statements as a variable, by default False.
        verbose : float, optional
            if True, printout statements printed automaically (ideal for use with iPython), by default True.
        """
        out = ""
        try:
            for idx, elevation, _, Px, Py, Mz in self.global_forces.itertuples(name=None):
                if any([Px, Py, Mz]):
                    string = f"\nLoad applied at elevation {elevation} m (node no. {idx}): Px = {Px} kN, Py = {Py} kN, Mx = {Mz} kNm."
                    if verbose is True:
                        print(string)
                    out += f"\nLoad applied at elevation {elevation} m (node no. {idx}): Px = {Px} kN, Py = {Py} kN, Mx = {Mz} kNm."
            if output is True:
                return out
        except Exception:
            print("No data found. Please create the Model first.")
            raise

    def set_pointload(
        self,
        elevation: float = 0.0,
        Py: float = None,
        Px: float = None,
        Mz: float = None,
    ):
        """
        Defines the point load(s) at a given elevation.

        .. note:
            If run several times at the same elevation, the loads are overwritten by the last command.


        Parameters
        ----------
        elevation : float, optional
            the elevation must match the elevation of a node, by default 0.0.
        Py : float, optional
            Shear force in kN, by default None.
        Px : float, optional
            Normal force in kN, by default None.
        Mz : float, optional
            Bending moment in kNm, by default None.
        """

        # identify if one node is at given elevation or if load needs to be split
        nodes_elevations = self.nodes_coordinates["x [m]"].values
        # check if corresponding node exist
        check = np.isclose(nodes_elevations, np.tile(elevation, nodes_elevations.shape), atol=0.001)

        try:
            if any(check):
                # one node correspond, extract node
                node_idx = int(np.where(check == True)[0])
                # apply loads at this node
                if Px is not None:
                    self.global_forces.loc[node_idx, "Px [kN]"] = Px
                if Py is not None:
                    self.global_forces.loc[node_idx, "Py [kN]"] = Py
                if Mz is not None:
                    self.global_forces.loc[node_idx, "Mz [kNm]"] = Mz
            else:
                if (
                    elevation > self.nodes_coordinates["x [m]"].iloc[0]
                    or elevation < self.nodes_coordinates["x [m]"].iloc[-1]
                ):
                    print(
                        "Load not applied! The chosen elevation is outside the mesh. The load must be applied on the structure."
                    )
                else:
                    print(
                        "Load not applied! The chosen elevation is not meshed as a node. Please include elevation in `x2mesh` variable when creating the Model."
                    )
        except Exception:
            print("\n!User Input Error! Please create Model first with the Model.create().\n")
            raise

    def set_pointdisplacement(
        self,
        elevation: float = 0.0,
        Ty: float = None,
        Tx: float = None,
        Rz: float = None,
    ):
        """
        Defines the displacement at a given elevation.

        .. note::
            for defining supports, this function should not be used, rather use `.set_support()`.

        Parameters
        ----------
        elevation : float, optional
            the elevation must match the elevation of a node, by default 0.0.
        Ty : float, optional
            Translation along y-axis, by default None.
        Tx : float, optional
            Translation along x-axis, by default None.
        Rz : float, optional
            Rotation around z-axis, by default None.
        """

        try:
            # identify if one node is at given elevation or if load needs to be split
            nodes_elevations = self.nodes_coordinates["x [m]"].values
            # check if corresponding node exist
            check = np.isclose(
                nodes_elevations, np.tile(elevation, nodes_elevations.shape), atol=0.001
            )

            if any(check):
                # one node correspond, extract node
                node_idx = int(np.where(check == True)[0])
                # apply displacements at this node
                if Tx is not None:
                    self.global_disp.loc[node_idx, "Tx [m]"] = Tx
                    self.global_restrained.loc[node_idx, "Tx"] = Tx > 0.0
                if Ty is not None:
                    self.global_disp.loc[node_idx, "Ty [m]"] = Ty
                    self.global_restrained.loc[node_idx, "Ty"] = Ty > 0.0
                if Rz is not None:
                    self.global_disp.loc[node_idx, "Rz [rad]"] = Rz
                    self.global_restrained.loc[node_idx, "Rz"] = Rz > 0.0
                # set restrain at this node

            else:
                if (
                    elevation > self.nodes_coordinates["x [m]"].iloc[0]
                    or elevation < self.nodes_coordinates["x [m]"].iloc[-1]
                ):
                    print(
                        "Support not applied! The chosen elevation is outside the mesh. The support must be applied on the structure."
                    )
                else:
                    print(
                        "Support not applied! The chosen elevation is not meshed as a node. Please include elevation in `x2mesh` variable when creating the Model."
                    )
        except Exception:
            print("\n!User Input Error! Please create Model first with the Model.create().\n")
            raise

    def set_support(
        self,
        elevation: float = 0.0,
        Ty: bool = False,
        Tx: bool = False,
        Rz: bool = False,
    ):
        """
        Defines the supports at a given elevation. If True, the relevant degree of freedom is restrained.

        .. note:
            If run several times at the same elevation, the support are overwritten by the last command.


        Parameters
        ----------
        elevation : float, optional
            the elevation must match the elevation of a node, by default 0.0.
        Ty : bool, optional
            Translation along y-axis, by default False.
        Tx : bool, optional
            Translation along x-axis, by default False.
        Rz : bool, optional
            Rotation around z-axis, by default False.
        """

        try:
            # identify if one node is at given elevation or if load needs to be split
            nodes_elevations = self.nodes_coordinates["x [m]"].values
            # check if corresponding node exist
            check = np.isclose(
                nodes_elevations, np.tile(elevation, nodes_elevations.shape), atol=0.001
            )

            if any(check):
                # one node correspond, extract node
                node_idx = int(np.where(check == True)[0])
                # apply loads at this node
                self.global_restrained.loc[node_idx, "Tx"] = Tx
                self.global_restrained.loc[node_idx, "Ty"] = Ty
                self.global_restrained.loc[node_idx, "Rz"] = Rz
            else:
                if (
                    elevation > self.nodes_coordinates["x [m]"].iloc[0]
                    or elevation < self.nodes_coordinates["x [m]"].iloc[-1]
                ):
                    print(
                        "Support not applied! The chosen elevation is outside the mesh. The support must be applied on the structure."
                    )
                else:
                    print(
                        "Support not applied! The chosen elevation is not meshed as a node. Please include elevation in `x2mesh` variable when creating the Model."
                    )
        except Exception:
            print("\n!User Input Error! Please create Model first with the Model.create().\n")
            raise

    def get_py_springs(self, kind: str = "node") -> pd.DataFrame:
        """Table with p-y springs computed for the given Model.

        Posible to extract the springs at the node level (i.e. spring at each node)
        or element level (i.e. top and bottom springs at each element)

        Parameters
        ----------
        kind : str
            can be of ("node", "element").

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with p-y springs, i.e. p-value [kN/m] and y-value [m].
        """
        if self.soil is None:
            return None
        else:
            if kind == "element":
                return misc.get_full_springs(
                    springs=self._py_springs,
                    elevations=self.nodes_coordinates["x [m]"].values,
                    kind="p-y",
                )
            elif kind == "node":
                return misc.get_reduced_springs(
                    springs=self._py_springs,
                    elevations=self.nodes_coordinates["x [m]"].values,
                    kind="p-y",
                )
            else:
                return None

    def get_mt_springs(self, kind: str = "node") -> pd.DataFrame:
        """Table with m-t (rotational) springs computed for the given Model.

        Posible to extract the springs at the node level (i.e. spring at each node)
        or element level (i.e. top and bottom springs at each element)

        Parameters
        ----------
        kind : str
            can be of ("node", "element").

        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with m-t springs, i.e. m-value [kNm] and t-value [-].
        """
        if self.soil is None:
            return None
        else:
            if kind == "element":
                return misc.get_full_springs(
                    springs=self._mt_springs,
                    elevations=self.nodes_coordinates["x [m]"].values,
                    kind="m-t",
                )
            elif kind == "node":
                return misc.get_reduced_springs(
                    springs=self._mt_springs,
                    elevations=self.nodes_coordinates["x [m]"].values,
                    kind="m-t",
                )
            else:
                return None

    def get_Hb_spring(self) -> pd.DataFrame:
        """Table with Hb (base shear) spring computed for the given Model.


        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with Hb spring, i.e. H-value [kN] and y-value [m].
        """
        if self.soil is None:
            return None
        else:
            spring_dim = self._Hb_spring.shape[-1]

            column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

            df = pd.DataFrame(
                data={
                    "Node no.": [self.element_number + 1] * 2,
                    "Elevation [m]": [self.pile.bottom_elevation] * 2,
                }
            )
            df["type"] = ["Hb", "y"]
            df[column_values_spring] = self._Hb_spring.reshape(2, spring_dim)

            return df

    def get_Mb_spring(self) -> pd.DataFrame:
        """Table with Mb (base moment) spring computed for the given Model.


        Returns
        -------
        pd.DataFrame (or None if no SoilProfile is present)
            Table with Mb spring, i.e. M-value [kNn] and t-value [-].
        """
        if self.soil is None:
            return None
        else:
            spring_dim = self._Hb_spring.shape[-1]

            column_values_spring = [f"VAL {i}" for i in range(spring_dim)]

            df = pd.DataFrame(
                data={
                    "Node no.": [self.element_number + 1] * 2,
                    "Elevation [m]": [self.pile.bottom_elevation] * 2,
                }
            )
            df["type"] = ["Mb", "t"]
            df[column_values_spring] = self._Hb_spring.reshape(2, spring_dim)

            return df

    def plot(self, assign=False):
        """Create a plot of the model with the mesh and boundary conditions.

        Parameters
        ----------
        assign : bool, optional
            this parameter can be set to True to return the figure, by default False.

        Returns
        -------
        matplotlib.pyplot.figure
            only return the object if assign=True.

        Examples
        --------

        *Plot without SoilProfile fed to the model:*

        .. image:: _static/plots/Model_no_soil_plot.png
           :scale: 70%

        *Plot with SoilProfile fed to the model:*

        .. image:: _static/plots/Model_with_soil_plot.png
           :scale: 70%
        """
        fig = graphics.connectivity_plot(self)
        return fig if assign else None

    @classmethod
    def create(
        cls,
        name: str,
        pile: Pile,
        soil: Optional[SoilProfile] = None,
        element_type: Literal["Timoshenko", "EulerBernoulli"] = "Timoshenko",
        x2mesh: List[float] = Field(default_factory=list),
        coarseness: float = 0.5,
        distributed_lateral: bool = True,
        distributed_moment: bool = True,
        distributed_axial: bool = False,
        base_shear: bool = True,
        base_moment: bool = True,
        base_axial: bool = False,
    ):
        """A method to create the Model.

        Parameters
        ----------
        name : str
            Name of the model
        pile : Pile
            Pile instance to be included in the model
        soil : Optional[SoilProfile], optional
            SoilProfile instance, by default None
        element_type : str, optional
            can be of ['EulerBernoulli','Timoshenko'], by default 'Timoshenko'
        x2mesh : List[float], optional
            additional elevations to be included in the mesh, by default none
        coarseness : float, optional
            maximum distance in meters between two nodes of the mesh, by default 0.5
        distributed_lateral : bool, optional
            include distributed lateral springs, by default True
        distributed_moment : bool, optional
            include distributed moment springs, by default False
        base_shear : bool, optional
            include lateral spring at pile toe, by default False
        base_moment : bool, optional
            include moment spring at pile toe, by default False

        Returns
        -------
        openpile.construct.Model
            a Model instance with a Pile structure and optionally a SoilProfile
        """

        obj = cls(
            name=name,
            pile=pile,
            soil=soil,
            element_type=element_type,
            x2mesh=x2mesh,
            coarseness=coarseness,
            distributed_lateral=distributed_lateral,
            distributed_moment=distributed_moment,
            distributed_axial=distributed_axial,
            base_shear=base_shear,
            base_moment=base_moment,
            base_axial=base_axial,
        )

        warnings.warn(
            "\nThe method Model.create() will be removed in version 1.0.0."
            "\nPlease use the base class to create a Pile instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return obj

    def __str__(self):
        return self.element_properties.to_string()



"""
`calculate` module
==================

The `calculate` module is used to run various functions outside the scope of the `analyze` module. 

Every function from this module returns an `openpile.compute.CalculateResult` object. 

"""

import pandas as pd
import numpy as np


class CalculateResult:
    _values: tuple


def bearingcapacity(
    model,
) -> CalculateResult:

    pass

    # # initiate dataframe
    # df = model.element_properties
    # # get vertical effective stress
    # sigveff = 0.5*(model.soil_properties["sigma_v top [kPa]"]+ model.soil_properties["sigma_v bottom [kPa]"])
    # elevation = 0.5*(model.soil_properties["x_top [m]"]+model.soil_properties[ "x_bottom [m]"])
    # # depth from ground
    # depth_from_ground = (0.5*(model.soil_properties["xg_top [m]"]+model.soil_properties["xg_bottom [m]"])).abs()
    # # pile width
    # pile_width = model.element_properties["Diameter [m]"]
    # # wall thickness
    # wall_thickness = model.element_properties["Wall thickness [m]"]

    # for layer in model.soil.layers:
    #     elements_for_layer = model.soil_properties.loc[
    #                 (model.soil_properties["x_top [m]"] <= layer.top)
    #                 & (model.soil_properties["x_bottom [m]"] >= layer.bottom)
    #             ].index

    #     if layer.axial_model is None:
    #         pass
    #     else:
    #         # Set local layer parameters for each element of the layer
    #         for i in elements_for_layer:
    #             pass



"""
`tz_curves` module
==================


"""

# Import libraries
import openpile.utils.misc as misc

import math as m
import numpy as np
from numba import njit, prange
from random import random

# Kraft et al (1989) formulation aid to ease up implementation
@njit(cache=True)
def kraft_modification(
    fmax: float,
    D: float,
    G0: float,
    residual: float = 1.0,
    tensile_factor: float = 1.0,
    RF: float = 0.9,
    zif: float = 10.0,
    output_length: int = 15,
):
    """
    Creates the t-z curve from relevant input with the Kraft et al (1981) formulation.

    Parameters
    ----------
    fmax: float
        unit skin friction [unit: kPa]
    D: float
        Pile diameter [unit: m]
    G0: float
        small-strain stiffness [unit: kPa]
    residual: float
        residual strength after peak strength
    tensile_factor: float
        strength factor for negative values of the curve
    RF: float
        curve fitting factor as per Kraft et al. (1981), by default 0.9
    zif: float
        dimensionless zone of influence as per Kraft et al (1981) that corresponds to the radius of the zone of influence divided by the pile radius, by default 10.0
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    """

    if output_length < 15:
        output_length = 15

    # define t till tmax
    tpos = np.linspace(0, fmax, output_length - 2)
    # define z till zmax
    zpos = tpos * 0.5 * D / G0 * np.log((zif - RF * tpos / fmax) / (1 - RF * tpos / fmax))
    # define z where t = tmax, a.k.a zmax here
    zmax = fmax * D / (2 * G0) * m.log((zif - RF) / (1 - RF))
    # define z where z=tres, which is zmax + 5mm
    zres = zmax + 0.005

    z = np.append(zpos, [zres, zres + 0.005])
    t = np.append(tpos, [residual * fmax, residual * fmax])

    return np.append(-z[-1::-1], np.append([0.0], z[1:])), np.append(
        -t[-1::-1] * tensile_factor, np.append([0.0], t[1:])
    )


# API clay function
@njit(cache=True)
def api_clay(
    sig: float,
    Su: float,
    D: float,
    residual: float = 0.9,
    tensile_factor: float = 1.0,
    output_length: int = 15,
):
    """
    Creates the API clay t-z curve from relevant input as per [API2000]_.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    Su : float
        Undrained shear strength [unit: kPa]
    D: float
        Pile diameter [unit: m]
    residual: float
        residual strength after peak strength, according to API-RP-2A,
        this value is between 0.7 and 0.9, default to 0.9
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    See also
    --------
    `API_clay`_

    """
    # cannot have less than 15
    if output_length < 15:
        output_length = 15

    # unit skin friction
    f = misc._fmax_api_clay(sig, Su)

    # piecewise function
    zlist = [0.0, 0.0016, 0.0031, 0.0057, 0.0080, 0.0100, 0.0200, 0.0300]
    tlist = [0.0, 0.3, 0.5, 0.75, 0.90, 1.00, residual, residual]

    # determine z vector
    z = np.array(zlist, dtype=np.float32) * D
    z = np.concatenate((-z[-1:0:-1], z))
    # define t vector
    t = np.array(tlist, dtype=np.float32) * f
    t = np.concatenate((-tensile_factor * t[-1:0:-1], t))

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_t_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = (0.02 + random() * 0.01) * D
        add_t_values[i] = residual * f

    z = np.append(z, add_z_values)
    t = np.append(t, add_t_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    t = t[z_id_sorted]

    return z, t


@njit(cache=True)
def api_clay_kraft(
    sig: float,
    Su: float,
    D: float,
    G0: float,
    residual: float = 1.0,
    tensile_factor: float = 1.0,
    RF: float = 0.9,
    zif: float = 10.0,
    output_length: int = 15,
):
    """
    Creates the API clay t-z curve (see [API2000]_) with the Kraft et al (1981) formulation (see [KrRK81]_).

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    Su : float
        Undrained shear strength [unit: kPa]
    D: float
        Pile diameter [unit: m]
    G0: float
        small-strain stiffness [unit: kPa]
    residual: float
        residual strength after peak strength, by default 1.0
    tensile_factor: float
        strength factor for negative values of the curve, by default 1.0
    RF: float
        curve fitting factor as per Kraft et al. (1981), by default 0.9
    zif: float
        dimensionless zone of influence as per Kraft et al (1981) that corresponds to the radius of the zone of influence divided by the pile radius, by default 10.0
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    See also
    --------
    `API_clay`_, :py:func:`openpile.utils.tz_curves.api_clay`

    """
    return kraft_modification(
        misc._fmax_api_clay(sig, Su), D, G0, residual, tensile_factor, RF, zif, output_length
    )


# API sand function
@njit(cache=True)
def api_sand(
    sig: float,
    delta: float,
    K: float = 0.8,
    tensile_factor: float = 1.0,
    output_length: int = 7,
):
    """
    Creates the API sand t-z curve (see [API2000]_).

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    delta: float
        interface friction angle [unit: degrees]
    K: float
        coefficient of lateral pressure (0.8 for open-ended piles and 1.0 for cloased-ended)
    tensile_factor: float
        strength factor for negative values of the curve
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 7, by default 7

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    See also
    --------
    `API_sand`_

    """
    # cannot have less than 7
    if output_length < 7:
        output_length = 7

    # unit skin friction
    f = misc._fmax_api_sand(sig, delta, K)

    # piecewise function
    zlist = [0.0, 0.0254, 0.03, 0.04]
    tlist = [0.0, 1.0, 1.0, 1.0]

    # determine z vector
    z = np.array(zlist, dtype=np.float32)
    z = np.concatenate((-z[-1:0:-1], z))
    # define t vector
    t = np.array(tlist, dtype=np.float32) * f
    t = np.concatenate((-tensile_factor * t[-1:0:-1], t))

    add_values = output_length - len(z)
    add_z_values = np.zeros((add_values), dtype=np.float32)
    add_t_values = np.zeros((add_values), dtype=np.float32)

    for i in range(add_values):
        add_z_values[i] = 0.03 + random() * 0.01
        add_t_values[i] = f

    z = np.append(z, add_z_values)
    t = np.append(t, add_t_values)

    z = np.sort(z)
    z_id_sorted = np.argsort(z)

    t = t[z_id_sorted]

    return z, t


@njit(cache=True)
def api_sand_kraft(
    sig: float,
    delta: float,
    D: float,
    G0: float,
    K: float = 0.8,
    residual: float = 1.0,
    tensile_factor: float = 1.0,
    RF: float = 0.9,
    zif: float = 10.0,
    output_length: int = 15,
):
    """
    Creates the API sand t-z curve (see [API2000]_) with the Kraft et al (1981) formulation (see [KrRK81]_).

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    delta: float
        interface friction angle [unit: degrees]
    D: float
        Pile diameter [unit: m]
    G0: float
        small-strain stiffness [unit: kPa]
    K: float
        coefficient of lateral pressure (0.8 for open-ended piles and 1.0 for cloased-ended), by default 0.8
    residual: float
        residual strength after peak strength, by default 1.0
    tensile_factor: float
        strength factor for negative values of the curve, by default 1.0
    RF: float
        curve fitting factor as per Kraft et al. (1981), by default 0.9
    zif: float
        dimensionless zone of influence as per Kraft et al (1981) that corresponds to the radius of the zone of influence divided by the pile radius, by default 10.0
    output_length : int, optional
        Number of discrete point along the springs, cannot be lower than 15, by default 15

    Returns
    -------
    numpy 1darray
        t vector [unit: kPa]
    numpy 1darray
        z vector [unit: m]

    See also
    --------
    `API_sand`_, :py:func:`openpile.utils.tz_curves.api_sand`

    """
    return kraft_modification(
        misc._fmax_api_sand(sig, delta, K), D, G0, residual, tensile_factor, RF, zif, output_length
    )



"""
`multipliers` module
====================



"""

import numpy as np


def durkhop(D: float, ra: float = 0.3) -> object:
    """This function generates p-multipliers as per Durkhop et al (see [Duhr09]_) for application
    to cyclic curves of the `openpile.soilmodels.API_sand` model.

    The default behaviour, i.e. when the parameter `ra` = 0.3 is that the API sand cyclic curves are unchanged
    with a equivalent number of cycles equal to 100.
    When the parameter `ra` is equal to 1.0, the cyclic curves are returned back to monotonic curves.
    In between 0.3 and 1.0, the multipliers may be thought as cyclic curves with a lower number of equivalent cycles.

    Parameters
    ----------
    D : float
        Pile diameter [m]
    ra : float, optional
        empirical factor dependent on the number of load cycles, 1.0 for monotonic loading and
        0.3 for cyclic loading at 100 cycles, by default 0.3

    Returns
    -------
    object
        callable for use as p_multipler in `openpile.soilmodels.API_sand` in cyclic mode.

    Example
    -------

    >>> from openpile.construct import Layer
    >>> from openpile.soilmodels import API_sand
    >>> from openpile.utils.multipliers import durkhop

    >>> # Create a Layer with API_sand and monotonic curves with Durkhop approach
    >>> Layer(
    >>>     name="medium dense sand",
    >>>     top=0,
    >>>     bottom=-40,
    >>>     weight=18,
    >>>     lateral_model=API_sand(
    >>>         phi=33,
    >>>         kind="cyclic",
    >>>         p_multiplier=durkhop(D=7.0, ra=1.0)
    >>>     ),
    >>> )


    Reference
    ---------

    .. [Duhr09] Dührkop, J. (2009). *Zum Einfluss von Aufweitungen und zyklischen Lasten auf
        das Verformungsverhalten lateral 385 beanspruchter Pfähle in Sand*. Ph.D. Thesis,
        Institut für Geotechnik und Baubetrieb, Technische Universität Hamburg-Harburg, Vol. 20 (in German).

    """

    func = lambda x: 1 / 0.9 * max(0.9, ra * (3 - 1.143 * x / D) + 0.343 * x / D)

    return func


def durkhop_normalized(D: float, ra: float = 0.3) -> object:
    """This function generates multipliers that represent ratios the cyclic and monotonic curves of
    the traditional API sand model.


    Parameters
    ----------
    D : float
        Pile diameter [m]
    ra : float, optional
        empirical factor dependent on the number of load cycles, 1.0 for monotonic loading and
        0.3 for cyclic loading at 100 cycles, by default 0.3

    Returns
    -------
    object
        callable for use as p_multipler in `openpile.soilmodels.LateralModel` or `openpile.soilmodels.AxialModel`
        The function input is the depth and the function output is the multiplier applied
        for the spring at the said depth.

    See also
    --------
    `durkhop`
    """

    func = lambda x: 0.9 / max(0.9, ra * (3 - 1.143 * x / D) + 0.343 * x / D)

    return func




from openpile import construct
from openpile.soilmodels import API_clay
import pytest
import numpy as np
import math as m

from pydantic import ValidationError
class TestPile:

    def test_main_constructor(self):
        """Main constructor of Pile object, check if pile.data 
        has even entries and that steel E value is 210GPa
        """
        # create a steel and circular pile
        pile = construct.Pile(kind='Circular',
                            name = "",
                            material='Steel',
                            top_elevation = 0,
                            pile_sections={
                                'length':[10,30],
                                'diameter':[7.5,8.5],
                                'wall thickness':[0.07, 0.08],
                            }
                            )
        # check Young modulus is indeed Steel
        assert pile.E == 210e6
        # check even numbers of row for dataframe
        assert pile.data.values.shape[0] % 2 == 0

    def test_pile_width(self):
        pile = construct.Pile(kind='Circular',
                            name = "",
                            material='Steel',
                            top_elevation = 0.1,
                            pile_sections={
                                'length':[20,20],
                                'diameter':[7.5,8.5],
                                'wall thickness':[0.07, 0.08],
                            }
                            )
        
        assert pile.width.mean() == 8.00

    def test_pile_length(self):
        pile = construct.Pile(kind='Circular',
                            name = "",
                            material='Steel',
                            top_elevation = 10,
                            pile_sections={
                                'length':[22,16],
                                'diameter':[7.5,8.5],
                                'wall thickness':[0.07, 0.08],
                            }
                            )
        
        assert pile.length == 38.0

    def test_pile_bottom(self):
        pile = construct.Pile(kind='Circular',
                            name = "",
                            material='Steel',
                            top_elevation = 10,
                            pile_sections={
                                'length':[10,30],
                                'diameter':[7.5,8.5],
                                'wall thickness':[0.07, 0.08],
                            }
                            )
        
        assert pile.bottom_elevation == -30.0


    def test_pile_area1(self):
        """check pile area 
        """
        # Create a pile instance with two sections of respectively 10m and 30m length.
        pile = construct.Pile(name = "",
                kind='Circular',
                material='Steel',
                top_elevation = 0,
                pile_sections={
                    'length':[10,30],
                    'diameter':[1.0,1.0],
                    'wall thickness':[0.5, 0.5],
                }
            )
        
        assert pile.area.mean() == 0.25*m.pi

    def test_pile_area2(self):
        pile = construct.Pile(name = "",
                kind='Circular',
                material='Steel',
                top_elevation = 0,
                pile_sections={
                    'length':[10,30],
                    'diameter':[1.0,1.0],
                    'wall thickness':[0.001, 0.001],
                }
            )
        
        assert m.isclose(pile.area.mean(), m.pi*0.001, rel_tol=0.01)

class TestLayer:
    def test_constructor(self):
        try:
            layer = construct.Layer(name='Soft Clay',
                            top=0,
                            bottom=-10,
                            weight=19,
                            lateral_model=API_clay(Su=[30,35], eps50=[0.01, 0.02], kind="cyclic"), 
                        )
        except Exception:
            assert False, "Constructor did not work"

    def test_wrong_order_top_and_bottom(self):
        with pytest.raises(ValidationError):
            layer = construct.Layer(name='Soft Clay',
                            top=-10,
                            bottom=0,
                            weight=19,
                            lateral_model=API_clay(Su=[30,35], eps50=[0.01, 0.02], kind="cyclic"), 
                        )

    def test_equal_top_and_bottom(self):
        with pytest.raises(ValidationError):
            layer = construct.Layer(name='Soft Clay',
                            top=-5,
                            bottom=-5,
                            weight=19,
                            lateral_model=API_clay(Su=[30,35], eps50=[0.01, 0.02], kind="cyclic"), 
                        )

    def test_wrong_unit_weight(self):
        """Constructor should not accept a soil unit weight equal to 10 or less as 
        it would be less dense than water and likely the user made a mistake with 
        the effective unit weight. 
        """
        with pytest.raises(ValidationError):
            layer = construct.Layer(name='Soft Clay',
                            top=0,
                            bottom=-10,
                            weight=10,
                            lateral_model=API_clay(Su=[30,35], eps50=[0.01, 0.02], kind="static"), 
                        )
            
class TestSoilProfile:
    def test_wrong_layers_elevations_gap(self):
        with pytest.raises(ValidationError):
            sp = construct.SoilProfile(
                name = "", 
                top_elevation=0, 
                water_line=0,
                layers=[
                    construct.Layer(name= "",
                                    top=0,
                                    bottom=-10,
                                    weight=20,
                                    ),
                    construct.Layer(name= "",
                                    top=-15,
                                    bottom=-30,
                                    weight=20,
                                    ),
                ])
    
    def test_wrong_layers_elevations_overlap(self):
        with pytest.raises(ValidationError):
            sp = construct.SoilProfile(
                name = "", 
                top_elevation=0, 
                water_line=0,
                layers=[
                    construct.Layer(name= "",
                                    top=0,
                                    bottom=-10,
                                    weight=20,
                                    ),
                    construct.Layer(name= "",
                                    top=-8,
                                    bottom=-30,
                                    weight=20,
                                    ),
                ])
            
    def test_wrong_top_elevation(self):
        with pytest.raises(ValidationError):
            sp = construct.SoilProfile(
                name = "", 
                top_elevation=10, 
                water_line=0,
                layers=[
                    construct.Layer(name= "",
                                    top=11,
                                    bottom=-10,
                                    weight=20,
                                    ),
                    construct.Layer(name= "",
                                    top=-10,
                                    bottom=-30,
                                    weight=20,
                                    ),
                ])
    
    def test_wrong_bottom_elevation(self):
        sp = construct.SoilProfile(
            name = "", 
            top_elevation=10, 
            water_line=0,
            layers=[
                construct.Layer(name= "",
                                top=10,
                                bottom=-10,
                                weight=20,
                                ),
                construct.Layer(name= "",
                                top=-10,
                                bottom=-30,
                                weight=20,
                                ),
            ])
    
        assert sp.bottom_elevation == -30

class TestModel:

    def test_constructor_wo_soil_models(self):
        """
        check that model can still be created with no lateral or axial models""" 
        try:
            model = construct.Model(
                name="",
                pile=construct.Pile(
                    name="",
                    top_elevation=20,
                    kind="Circular",
                    material="Steel",
                    pile_sections={
                        'length':[50],
                        'diameter':[7],
                        'wall thickness':[0.08]
                    }
                ),
                soil = construct.SoilProfile(
                    name = "", 
                    top_elevation=10, 
                    water_line=0,
                    layers=[
                        construct.Layer(name= "",
                                        top=10,
                                        bottom=-10,
                                        weight=20,
                                        ),
                        construct.Layer(name= "",
                                        top=-10,
                                        bottom=-30,
                                        weight=20,
                                        )]
                ),
                )
        except Exception:
            assert False, f"Constructor does not work"
        
        # check that p_y springs are all zero
        assert not np.all( model.get_py_springs()[['VAL 0','VAL 1','VAL 2']].values )

        # # check that m_t springs are all zero
        # assert not np.all( model.get_mt_springs()[['VAL 0','VAL 1','VAL 2']].values )


"""
`py_curves` module
==================

"""

# Import libraries
import math as m
import numpy as np
from numba import njit, prange
from random import random

from openpile.core.misc import conic

# SPRING FUNCTIONS --------------------------------------------
@njit(cache=True)
def cowden_clay(
    X: float,
    Su: float,
    G0: float,
    D: float,
    output_length: int = 20,
):
    """
    Creates the lateral springs from the PISA clay formulation
    published by Byrne et al 2020 (see [BHBG20]_) and calibrated based pile
    load tests at Cowden (north east coast of England).

    Parameters
    ----------
    X : float
        Depth below ground level [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    """

    # # Cowden clay parameters
    v_pu = 241.4
    k_p1 = 10.6
    k_p2 = -1.650
    n_p1 = 0.9390
    n_p2 = -0.03345
    p_u1 = 10.7
    p_u2 = -7.101

    # Depth variation parameters
    v_max = v_pu
    k = k_p1 + k_p2 * X / D
    n = n_p1 + n_p2 * X / D
    p_max = p_u1 + p_u2 * m.exp(-0.3085 * X / D)

    # calculate normsalised conic function
    y, p = conic(v_max, n, k, p_max, output_length)

    # return non-normalised curve
    return y * (Su * D / G0), p * (Su * D)


@njit(cache=True)
def dunkirk_sand(
    sig: float,
    X: float,
    Dr: float,
    G0: float,
    D: float,
    L: float,
    output_length: int = 20,
):
    """
    Creates the lateral spring from the PISA sand formulation
    published  by Burd et al (2020) (see [BTZA20]_).
    Also called the General Dunkirk Sand Model (GDSM).

    Parameters
    ----------
    sig : float
        vertical/overburden effective stress [unit: kPa]
    X : float
        Depth below ground level [unit: m]
    Dr : float
        Sand relative density Value must be between 0 and 100 [unit: -]
    G0 : float
        Small-strain shear modulus [unit: kPa]
    D : float
        Pile diameter [unit: m]
    L : float
        Embedded pile length [unit: m]
    output_length : int, optional
        Number of datapoints in the curve, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]
    """
    # correct relative density for decimal value
    Dr = Dr / 100

    # Generalised Dunkirk Sand Model parameters
    v_pu = 146.1 - 92.11 * Dr
    k_p1 = 8.731 - 0.6982 * Dr
    k_p2 = -0.9178
    n_p = 0.917 + 0.06193 * Dr
    p_u1 = 0.3667 + 25.89 * Dr
    p_u2 = 0.3375 - 8.9 * Dr

    # Depth variation parameters
    v_max = v_pu
    k = k_p1 + k_p2 * X / D
    n = n_p
    p_max = p_u1 + p_u2 * X / L

    # calculate normsalised conic function
    y, p = conic(v_max, n, k, p_max, output_length)

    # return non-normalised curve
    return y * (sig * D / G0), p * (sig * D)


# API sand function
@njit(parallel=True, cache=True)
def api_sand(
    sig: float,
    X: float,
    phi: float,
    D: float,
    kind: str = "static",
    below_water_table: bool = True,
    ymax: float = 0.0,
    output_length: int = 20,
):
    """
    Creates the API sand p-y curve from relevant input.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    X: float
        Depth of the curve w.r.t. mudline [unit: m]
    phi: float
        internal angle of friction of the sand layer [unit: degrees]
    D: float
        Pile width [unit: m]
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    below_water_table: bool, by default False
        switch to calculate initial subgrade modulus below/above water table
    ymax: float, by default 0.0
        maximum value of y, default goes to 99.9% of ultimate resistance
    output_length: int, by default 20
        Number of discrete point along the springs

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]
    """
    # A value - only thing that changes between cyclic or static
    if kind == "static":
        A = max(0.9, 3 - 0.8 * X / D)
    else:
        A = 0.9

    # initial subgrade modulus in kpa from fit with API (2014)
    if below_water_table:
        k_phi = max((0.1978 * phi**2 - 10.232 * phi + 136.82) * 1000, 5400)
    else:
        k_phi = max((0.2153 * phi**2 - 8.232 * phi + 63.657) * 1000, 5400)

    # Calculate Pmax (regular API)
    ## Factors according to Mosher and Dawkins 2008.(regular API)
    b = 0.4
    Beta = 45 + phi / 2
    rad = m.pi / 180
    C1 = (
        (b * m.tan(phi * rad) * m.sin(Beta * rad))
        / (m.tan((Beta - phi) * rad) * m.cos((phi / 2) * rad))
        + ((m.tan(Beta * rad)) ** 2 * m.tan((phi / 2) * rad)) / (m.tan((Beta - phi) * rad))
        + b * m.tan(Beta * rad) * (m.tan(phi * rad) * m.sin(Beta * rad) - m.tan((phi / 2) * rad))
    )
    C2 = m.tan(Beta * rad) / m.tan((Beta - phi) * rad) - (m.tan((45 - phi / 2) * rad)) ** 2
    C3 = b * m.tan(phi * rad) * (m.tan(Beta * rad)) ** 4 + (m.tan((45 - phi / 2) * rad)) ** 2 * (
        (m.tan(Beta * rad)) ** 8 - 1
    )

    ## Pmax for shallow and deep zones (regular API)
    Pmax = min(C3 * sig * D, C1 * sig * X + C2 * sig * D)

    # creation of 'y' array
    if ymax == 0.0:
        # ensure X cannot be zero
        z = max(X, 0.1)
        if Pmax == 0:
            f = (C3 * D + C2 * D) / 2
        else:
            f = Pmax
        ymax = 4 * A * f / (k_phi * z)

    # determine y vector from 0 to ymax
    # y_ini = np.array([0,],dtype=np.float64)
    # y = np.append(y_ini,np.logspace(np.log10(0.0001*D),np.log10(ymax),output_length-1))
    y = np.linspace(0, ymax, output_length)

    ## calculate p vector
    p = np.zeros(shape=len(y), dtype=np.float32)
    for i in prange(len(y)):
        if Pmax == 0:
            p[i] = 0
        else:
            p[i] = A * Pmax * m.tanh((k_phi * X * y[i]) / (A * Pmax))

    return y, p


# API clay function
@njit(parallel=True, cache=True)
def api_clay(
    sig: float,
    X: float,
    Su: float,
    eps50: float,
    D: float,
    J: float = 0.5,
    stiff_clay_threshold=96,
    kind: str = "static",
    ymax: float = 0.0,
    output_length: int = 20,
):
    """
    Creates the API clay p-y curve from relevant input.

    Parameters
    ----------
    sig: float
        Vertical effective stress [unit: kPa]
    X: float
        Depth of the curve w.r.t. mudline [unit: m]
    Su : float
        Undrained shear strength [unit: kPa]
    eps50: float
        strain at 50% ultimate resistance [-]
    D: float
        Pile width [unit: m]
    J: float, by default 0.5
        empirical factor varying depending on clay stiffness
    stiff_clay_threshold: float, by default 96.0
        undrained shear strength at which stiff clay curve is computed [unit: kPa]
    kind: str, by default "static"
        types of curves, can be of ("static","cyclic")
    ymax: float, by default 0.0
        maximum value of y, if null the maximum is calculated such that the whole curve is computed
    output_length: int, by default 20
        Number of discrete point along the springs

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]
    """
    # important variables
    y50 = 2.5 * eps50 * D
    if X == 0.0 or Su == 0:
        Xr = 2.5 * D
    else:
        Xr = max((6 * D) / (sig / X * D / Su + J), 2.5 * D)

    # creation of 'y' array
    if ymax == 0.0:
        ymax = 16 * y50

    # Calculate Pmax (regular API)
    ## Pmax for shallow and deep zones (regular API)
    Pmax_shallow = (3 * Su + sig) * D + J * Su * X
    Pmax_deep = 9 * Su * X
    Pmax = min(Pmax_deep, Pmax_shallow)

    ylist_in = [0.0, 0.1 * y50, 0.21 * y50, 1 * y50, 3 * y50, 8 * y50, 15 * y50, ymax]
    ylist_out = []
    for i in range(len(ylist_in)):
        if ylist_in[i] <= ymax:
            ylist_out.append(ylist_in[i])

    # determine y vector from 0 to ymax
    y = np.array(ylist_out, dtype=np.float32)
    add_values = output_length - len(y)
    add_y_values = []
    for _ in range(add_values):
        add_y_values.append(0.1 * y50 + random() * (ymax - 0.1 * y50))
    y = np.append(y, add_y_values)
    y = np.sort(y)

    # define p vector
    p = np.zeros(shape=len(y), dtype=np.float32)

    for i in prange(len(y)):
        if kind == "static":
            # derive static curve
            if y[i] > 8 * y50:
                p[i] = Pmax
            else:
                p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33
        else:
            # derive cyclic curve
            if X <= Xr:
                if Su <= stiff_clay_threshold:
                    if y[i] > 15 * y50:
                        p[i] = 0.7185 * Pmax * X / Xr
                    elif y[i] > 3 * y50:
                        p[i] = 0.7185 * Pmax * (1 - (1 - X / Xr) * (y[i] - 3 * y50) / (12 * y50))
                    else:
                        p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33
                else:
                    if y[i] > 15 * y50:
                        p[i] = 0.5 * Pmax * X / Xr
                    elif y[i] > 1 * y50:
                        p[i] = 0.5 * Pmax * (1 - (1 - X / Xr) * (y[i] - 1 * y50) / (14 * y50))
                    else:
                        p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

            elif X > Xr:
                if Su <= stiff_clay_threshold:
                    if y[i] > 3 * y50:
                        p[i] = 0.7185 * Pmax
                    else:
                        p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33
                else:
                    if y[i] > 1 * y50:
                        p[i] = 0.5 * Pmax
                    else:
                        p[i] = 0.5 * Pmax * (y[i] / y50) ** 0.33

        # modification of initial slope of the curve (DNVGL RP-C203 B.2.2.4)
        if y[i] == 0.1 * y50:
            p[i] = 0.23 * Pmax

    return y, p


@njit(parallel=True, cache=True)
def reese_weakrock(
    Ei: float,
    qu: float,
    RQD: float,
    xr: float,
    D: float,
    k: float = 0.0005,
    output_length=20,
):
    """creates the Reese weakrock p-y curve based on the work of Reese (1997) #TODO(ref).


    Parameters
    ----------
    Ei : float
        initial modulus of rock [kPa]
    qu : float
        compressive strength of rock [kPa]
    RQD : float
        Rock Quality Designation [%]
    xr : float
        depth from rock surface [m]
    D : float
        pile width [m]
    k : float, optional
        dimensional constant, randing from 0.0005 to 0.00005, by default 0.0005
    output_length : int, optional
        length of output arrays, by default 20

    Returns
    -------
    1darray
        y vector [unit: m]
    1darray
        p vector [unit: kN/m]

    """

    # Rqd forced to be within 0 and 100
    rqd = max(min(100, RQD), 0)

    # determine alpha
    alpha = 1.0 - 2 / 3 * rqd / 100

    # determine ultimate resistance of rock
    Pmax = min(5.2 * alpha * qu * D, alpha * qu * D * (1 + 1.4 * xr / D))

    # initial portion of p-y curve
    Epy_i = Ei * min(500, 100 + 400 * xr / (3 * D))

    # yA & yrm
    yrm = k * D
    yA = (Pmax / (2 * (yrm) ** 0.25 * Epy_i)) ** 1.333

    # define y
    ymax = max(1.05 * yA, (2 * yrm ** (0.25)) ** 4)
    y = np.linspace(yA, ymax, output_length - 2)
    y = np.concatenate((np.array([0.0]), y, np.array([1.2 * ymax])))

    # define p
    p = np.zeros(y.size)
    for i in range(len(p)):
        if y[i] <= yA:
            p[i] = min(Pmax, Epy_i * y[i])
        else:
            p[i] = min(Pmax, Pmax / 2 * (y[i] / yrm) ** 0.25)

    return y, p
